//////////////////////////////////////////////////////////////////////////////////
//
//  handlers.s 
//
//  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
//
//  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
//
//////////////////////////////////////////////////////////////////////////////////
//
// $Author: jmaebe $
// $Date: 2002/03/22 14:31:27 $
// $Log: handlers.s,v $
// Revision 1.4.2.13  2002/03/22 14:31:27  jmaebe
// * the return address of indeirect calls to libpthread routines is now
//   looked up using LOOKUP_RET_ADDRESS (in the past, only (d)delta was added)
//
// Revision 1.4.2.12  2002/03/20 10:20:58  jmaebe
// + jmpcall_dir_handler_ (forgot to commit last time)
//
// Revision 1.4.2.11  2002/03/11 21:21:59  jmaebe
// * MAKE_LINK is now no longer used while generating code, since it also
//   caused problems when using MT code (thread ! is instrumenting, generats a
//   link to its instrumented code which isn't finished yet, thread 2 returns
//   from a function or jumps/calls to the linked address and starts executing
//   the unfinished instrumented code) -> changed to SCHEDULE_MAKE_LINK() and
//   committed after the current block is finished
// * LOOKUP_CLONE2INSTR() once again uses the jumps in the clone to determine
//   targets
// * libpthread is no longer instrumented because it disables all signals
//   (using syscall 0xb3) in some routines, causing JiTI to hang (since its
//   breakpoints aren't processed anymore, it can't continue instrumenting and
//   the program can't continue executing)
// * made jcc_taken_handler() thread-safe: the MAKE_LINK() there was not safe
//   since it consists of 2 PUT_****() operations, so another thread could be
//   executing the original call while it was being replaced with a JMP. At
//   the same time, PATCH_INFO was made obsolete by doing the patching inside
//   jcc_taken_handler (and that should be more or less thread safe, since it
//   replaces exactly one int) -> GEBRUIK_PATCH_INFO is always set to 0 in
//   jiti.c, can possibly be removed.
// + infrastructure for intercepting syscalls and modifying the values passed
//   to the kernel, not used anymore (I used it to debug the syscall 0xb3
//   problem)
// * separate assembler handlers for adjusting the target of a jmp or call
//   mem/reg when instrumenting parallel programs (the call handler checks
//   for targets in libpthread, the jump handler doesn't). The common handler
//   is still used if we aren't instrumenting a parallel program (and that
//   one obviously doesn't include the pthread checks)
// * the memory layout is now always written to the log
// - commented out some unused global variables in instrument.c
// * some cosmetic changes and translations
// * with the above changes, ploop now always works (on a dual proc machine)
//   with the opt and trace-mem backends). ptest now always crashed for some
//   reason however. If you use trace-code, you also almost always get crashes
//   with MT programs.
// And that's all for today, folks!
//
// Revision 1.4.2.10  2002/03/08 15:17:24  jmaebe
// + added code to stop instrumenting when "exit" is called again (it got lost
//   when adding the lookup of jmp/call mem/reg addresses)
//
// Revision 1.4.2.9  2002/03/07 17:26:22  jmaebe
// * jmp/call mem/reg targets are now also looked up in clone2instr etc (using
//   the LOOKUP_RET_HANDLER())
// * helpers foor looking up targets and jcc_taken_handler in handlers.s now
//   only save registers that are not saved by the standard calling conventions
//   (eax, ecx and edx)
// * moved code to add a target to the clone2instr table to a separate procedure
// * increased clone2instr cache size to 4MB (1M entries)
// * when following direct calls, the entry points of those procedures are
//   added now to clone2instr as well. Doing the same for jumps seemed to only
//   slow down everything.
// * total speed improvement (for FPC): 27%
// * decrease in size of generated instrumented code (for FPC): 50%
//
// Revision 1.4.2.8  2002/02/25 16:25:17  jmaebe
// + deactivate jiti when exit() is called too (fixes crash at the end of gzip)
//
// Revision 1.4.2.7  2002/02/22 17:03:42  jmaebe
// * made checking of ret from main thread safe
// * LOOKUP_RET_HANDLER now also uses LOOKUP_CLONE2INSTR and LOOKUP_RELOC when
//   the ret address is in a dynlib
// * ret_handler (handlers.s) now saves all registers (it calls the C-routine
//   LOOKUP_RET_HANDLER), should be slightly faster and is now jumped at
//   instead of called
// * fac now works reliably to when instrumented!!
//
// Revision 1.4.2.6  2002/02/22 14:23:31  jmaebe
// + save return address of main at the start of the program and use it to
//   stop instrumenting when we return from main
//
// Revision 1.4.2.5  2002/02/20 12:20:50  jmaebe
// - removed unnecessary check for fini2 in adjust_target_
//
// Revision 1.4.2.4  2002/02/14 17:47:32  jmaebe
// + support for "lock" (merged, including fix)
// + support for cmpxchg (merged)
// * made jmp/call reg/mem again even slightly faster :)
// + support for in/out
//
// Revision 1.4.2.3  2002/02/13 15:48:25  jmaebe
// * optmized jmp_mem/call_mem and jmp_reg/call_reg handling
// * fixed interrupt_handler()
//
// Revision 1.4.2.2  2002/02/13 12:16:58  jmaebe
// + support for interrupts
// * instrumenting dynamic libraries now works, but libc still causes some
//   problems because of unsupported instructions (e.g lock and cmpxchgl)
//
// Revision 1.4.2.1  2002/02/12 16:47:20  jmaebe
// + initial implementation of instrumenting of dynamically loaded libraries,
//   but not yet working (even main(void) {} crashes :). Also, some backends
//   which have to be modified before they can work with this new feature
//   haven't been adapted yet. Compiler with -DINSTRUMENT_DYNLIBS to include
//   the functionality (it's even required currently, because there's a
//   reference to ddelta (dynlib delta) in handlers.s which can't be disabled
//   using a conditional. Method used: separate clone for the dynlibs (thanks
//   for the idea, Michiel :)
//
// Revision 1.4  2001/11/14 14:04:49  ronsse
// Changed /* to // type of comments in copyright notice.
//
// Revision 1.3  2001/10/25 13:46:01  ronsse
// Initial commit
//
//
//////////////////////////////////////////////////////////////////////////////////


.include "macros.inc"
	
.text
        .align 4

/////////////////////////////////////////////////////////////////////////////////////
                
.globl get_bp
        .type get_bp,@function
get_bp: 
        movl %ebp,%eax
        ret
        .size get_bp,.-get_bp
	
/////////////////////////////////////////////////////////////////////////////////////

// NOTE: this routine expects to be *jumped* at, not called!!!
.globl adjust_target_no_pthread_
	.type adjust_target_no_pthread_, @function
adjust_target_no_pthread_:

	// this handler is *jumped* at, not called -> argument is now at (%esp)	
	pushl  %eax
	pushl  %ecx
	pushl  %edx
	movl   12(%esp), %eax
	pushf
	// NOW:	 eax=jump address

	// check for things we can't instrument

	// 1) when we go to exit, stop (since that will at one point also
	//    cause jiti to be unloaded and its logging files to be closed)
	cmpl  exit_jmpaddr, %eax
	je   .Lno_pthread_noadjust

	// check where we have to return to
	pushl %eax
	call LOOKUP_RET_ADDRESS

	// remove LOOKUP_RET_ADDRESS argument (pushed eax) from the stack
	leal 4(%esp), %esp
	
	// put modified address back on the stack (there has been a extra pushf)
	movl %eax, 16(%esp)

.Lno_pthread_noadjust:
	// clean up and return/jump to the (new) target
	popf
	popl %edx
	popl %ecx
	popl %eax
	ret
        .size adjust_target_no_pthread_,.-adjust_target_no_pthread_

/////////////////////////////////////////////////////////////////////////////////////

// NOTE: this routine expects to be *jumped* at, not called!!!
.globl adjust_call_target_
	.type adjust_call_target_, @function
adjust_call_target_:

	// this handler is *jumped* at, not called -> argument is now at (%esp)	
	pushl  %eax
	pushl  %ecx
	pushl  %edx
	movl   12(%esp), %eax
	pushf
	// NOW:	 eax=jump address

	// check for things we can't instrument

	// 1) when we go to exit, stop (since that will at one point also
	//    cause jiti to be unloaded and its logging files to be closed)
	cmpl  exit_jmpaddr, %eax
	je   .Lcallreg_noadjust

	// 2) don't instrument libpthread. The reason is that libpthread
	//    disables signals in some places, causing the instrumented
	//    program to hang since jiti can't handle the breakpoints anymore
	cmpl libpthread_base,%eax
	jb   .Lno_pthread
	cmpl libpthread_top,%eax
	movl 20(%esp),%edx
	// improve jump prediction
	jbe   .Lpthread
	jmp   .Lno_pthread
.Lpthread:
	// call to libpthread address -> adjust returnaddress so it points
	// to the clone and call the routine directly
	pushl %edx
	call LOOKUP_RET_ADDRESS
	leal 4(%esp), %esp
	movl %eax, 20(%esp)

	jmp  .Lcallreg_noadjust
.Lno_pthread:
	
	// check where we have to return to
	pushl %eax
	call LOOKUP_RET_ADDRESS

	// remove LOOKUP_RET_ADDRESS argument (pushed eax) from the stack
	leal 4(%esp), %esp
	
	// put modified address back on the stack (there has been a extra pushf)
	movl %eax, 16(%esp)

.Lcallreg_noadjust:
	// opkuisen & springen
	popf
	popl %edx
	popl %ecx
	popl %eax
	ret
        .size adjust_call_target_,.-adjust_call_target_

/////////////////////////////////////////////////////////////////////////////////////

// NOTE: this routine expects to be *jumped* at, not called!!!
.globl adjust_jump_target_
	.type adjust_jump_target_, @function
adjust_jump_target_:

	// this handler is *jumped* at, not called -> argument is now at (%esp)	
	pushl  %eax
	pushl  %ecx
	pushl  %edx
	movl 12(%esp), %eax
	pushf

	// NOW:	 eax=jump address
//	cmpl  exit_jmpaddr, %eax
//	je    .Ljmpreg_noadjust

	// check where we have to return to
	pushl %eax
	call LOOKUP_RET_ADDRESS

	// remove LOOKUP_RET_ADDRESS argument (pushed eax) from the stack
	leal 4(%esp), %esp
	
	// put modified address back on the stack (there has been a extra pushf)
	movl %eax, 16(%esp)

//.Ljmpreg_noadjust:
	// opkuisen & springen
	popf
	popl %edx
	popl %ecx
	popl %eax
	ret
        .size adjust_jump_target_,.-adjust_jump_target_

/////////////////////////////////////////////////////////////////////////////////////

.globl jmpcall_dir_handler_
	.type jmpcall_dir_handler_, @function

jmpcall_dir_handler_:

	pushl  %eax
	pushl  %ecx
	pushl  %edx
	pushf	

	movl 16(%esp), %eax
	// arg=IP
	pushl %eax
	call jmpcall_dir_handler
	leal 4(%esp), %esp
	movl %eax, 16(%esp)
	
	popf
	popl  %edx
	popl  %ecx
	popl  %eax

	ret
	
	.size jmpcall_dir_handler_,.-jmpcall_dir_handler_
	
/////////////////////////////////////////////////////////////////////////////////////

.globl jcc_taken_handler_
	.type jcc_taken_handler_, @function

jcc_taken_handler_:

	pushl  %eax
	pushl  %ecx
	pushl  %edx
	pushf	

	movl 16(%esp), %eax
	// arg=IP
	pushl %eax
	call jcc_taken_handler
	leal 4(%esp), %esp
	movl %eax, 16(%esp)
	
	popf
	popl  %edx
	popl  %ecx
	popl  %eax

	ret
	
	.size jcc_taken_handler_,.-jcc_taken_handler_
	
/////////////////////////////////////////////////////////////////////////////////////

.globl interrupt_modifier_
	.type interrupt_handler_, @function

interrupt_modifier_:

	pushl  %ebp
	movl   %esp,%ebp
	pushl  %eax
	pushl  %ecx
	pushl  %edx
	pushf	

	pushl %edx
	pushl %ecx
	pushl %ebx
	pushl %eax
	call interrupt_modifier
	addl $16,%esp

	popf
	popl  %edx
	popl  %ecx
	popl  %eax
	leave
	ret
	
	.size interrupt_modifier_,.-interrupt_modifier_

/////////////////////////////////////////////////////////////////////////////////////	
		
		
.globl ret_handler_
	.type ret_handler_, @function

ret_handler_:
	
	// this handler is *jumped* at, not called -> argument is now at (%esp)	
	pushl  %eax
	pushl  %ecx
	pushl  %edx
	movl 12(%esp), %eax
	pushf

	// NOW:	 eax=jump address

	// stop instrumenting when we return from main
	cmpl main_ret,%eax
	je  .LNoAdjust

	// check where we have to return to
	pushl %eax
	call LOOKUP_RET_ADDRESS

	// remove LOOKUP_RET_ADDRESS argument (pushed eax) from the stack
	leal 4(%esp), %esp
	
	// put modified address back on the stack (there has been a extra pushf)
	movl %eax, 16(%esp)
.LNoAdjust:
	// opkuisen & springen
	popf
	popl %edx
	popl %ecx
	popl %eax	
	ret
	
	.size ret_handler_,.-ret_handler_

	
/////////////////////////////////////////////////////////////////////////////////////

		
.globl trace_
        .type trace_,@function
	
trace_:
	
    //  reti 16
	
        pushl %ebp
        movl %esp, %ebp
        
        pushal
        pushf

	// 4 argumenten kopieren
	
	//  ADDRESS	
        movl 20(%ebp), %eax
        pushl %eax
	
	//  LEN
        movl 16(%ebp), %eax
        pushl %eax
	
	//  L/S
        movl 12(%ebp), %eax
        pushl %eax
	
	//  IP+delta+1
        movl 8(%ebp), %eax
        pushl %eax
	
        call *CB_data
        addl $16, %esp	
        popf
        popal
        leave
	
	// cleanup!!
        reti 16
	
        .size trace_,.-trace_

//////////////////////////////////////////////////////////////////


.globl trace_code_
        .type trace_code_,@function
	
trace_code_:   //  trace_code_(start, stop)
        
//      reti 8
	
        pushl %ebp
        movl %esp, %ebp
        
        pushal
        pushf

        // 2 argumenten kopieren
	

        movl 8(%ebp), %eax
        pushl %eax
        movl 12(%ebp), %eax
        pushl %eax
        
        call *CB_code
        addl $8, %esp
        popf
        popal
        leave
	
	// cleanup!!
        reti 8

        .size trace_code_,.-trace_code_

