/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  jiti.h 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: jmaebe $
 * $Date: 2002/03/26 13:15:53 $
 * $Log: jiti.h,v $
 * Revision 1.8.2.12  2002/03/26 13:15:53  jmaebe
 * * fixed detection of PLT jmp_mem entries
 * * if we detect that we are going to return to a breakpoint in
 *   LOOKUP_RET_ADDRESS, we immediately instrument that code
 * * the jump coming after instrumented "int 0x80"'s is now replcaced with a
 *   a call to jmpcall_dir_handler too
 * * the clone wasn't refilled partially with breakpoints when we followed a
 *   direct short jump (with instruction length 2)
 *  -> with the 3 previous changes, we now only touch breakpoints when starting
 *   the program and when returning from the first call to a routine in
 *   libpthread (-> FPC has only one breakpoint left, pdf2ps 7)
 *
 * Revision 1.8.2.11  2002/03/12 09:15:42  jmaebe
 * * removed some warnings
 *
 * Revision 1.8.2.10  2002/03/08 20:09:48  jmaebe
 * + declaration of get_exit_jmpaddr()
 *
 * Revision 1.8.2.9  2002/03/07 17:26:22  jmaebe
 * * jmp/call mem/reg targets are now also looked up in clone2instr etc (using
 *   the LOOKUP_RET_HANDLER())
 * * helpers foor looking up targets and jcc_taken_handler in handlers.s now
 *   only save registers that are not saved by the standard calling conventions
 *   (eax, ecx and edx)
 * * moved code to add a target to the clone2instr table to a separate procedure
 * * increased clone2instr cache size to 4MB (1M entries)
 * * when following direct calls, the entry points of those procedures are
 *   added now to clone2instr as well. Doing the same for jumps seemed to only
 *   slow down everything.
 * * total speed improvement (for FPC): 27%
 * * decrease in size of generated instrumented code (for FPC): 50%
 *
 * Revision 1.8.2.8  2002/03/06 20:54:55  jmaebe
 * * sped up jiti by about 1000% by using a hashing function to lookup entries
 *   in clone2instr. FPC now runs only about 4 times slower with instrumentation
 *   than without.
 *
 * Revision 1.8.2.7  2002/03/06 12:49:23  jmaebe
 * * improved LOOKUP_CLONE2INSTR significantly
 * * corrected behaviour of FOLLOW_JUMP/CALL_DIR (we followed jumps/calls if
 *   those vars were false instead of true
 * * adapted the JMPL macro so that LOOKUP_CLONE2INSTR remains thread-safe
 *
 * Revision 1.8.2.6  2002/02/28 15:53:34  jmaebe
 * * signal handlers are re-entrant under Linux nowadays, but bp_handlet set/
 *   used a global variable, causing general havoc with multi-threaded programs
 *   -> bp changed to a local variable (it isn't used outside bp_handler anyway)
 * * __test_and_set() prototype moved from insturment.c to jiti.h
 * * clone2instr updating now happens after the current block has been finished,
 *   otherwise a thread could start executing the unfinished block of
 *   instructions that another thread is generating
 * * interrupt_handler() only sets X_kloon to 1 if we aren't instrumenting code
 * * removed unnecessary locks in jcc_taken_handler()
 *
 * !!WARNING: don't ever use pthread_mutex_lock() etc, because these functions
 *   don't seem to be re-entrant so we hang if we call them while we are
 *   instrumenting them.
 *
 * Revision 1.8.2.5  2002/02/25 16:25:17  jmaebe
 * + deactivate jiti when exit() is called too (fixes crash at the end of gzip)
 *
 * Revision 1.8.2.4  2002/02/22 14:48:35  jmaebe
 * * fixed gs override support
 *
 * Revision 1.8.2.3  2002/02/22 14:23:31  jmaebe
 * + save return address of main at the start of the program and use it to
 *   stop instrumenting when we return from main
 *
 * Revision 1.8.2.2  2002/02/20 08:30:51  jmaebe
 * - removed 'ifdef/endif INSTRUMENT_DYNLIBS" defines
 * + support for "ret imm" (needed for Pascal code :) (merged r1.9)
 * + all 32bit registers defined (only EAX was defined until now) (merged r1.10)
 *
 * Revision 1.8.2.1  2002/02/12 16:47:20  jmaebe
 * + initial implementation of instrumenting of dynamically loaded libraries,
 *   but not yet working (even main(void) {} crashes :). Also, some backends
 *   which have to be modified before they can work with this new feature
 *   haven't been adapted yet. Compiler with -DINSTRUMENT_DYNLIBS to include
 *   the functionality (it's even required currently, because there's a
 *   reference to ddelta (dynlib delta) in handlers.s which can't be disabled
 *   using a conditional. Method used: separate clone for the dynlibs (thanks
 *   for the idea, Michiel :)
 *
 * Revision 1.8  2002/02/05 17:20:29  jmaebe
 *   * PUSHL8 fixed (argument is a char, not a long)
 *
 * Revision 1.7  2002/01/10 21:25:30  ronsse
 * *** empty log message ***
 *
 * Revision 1.6  2001/11/12 10:11:30  ronsse
 * Added jiti_lock/jiti_unlock for backends
 *
 * Revision 1.5  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * Common include file for JiTI86.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#ifndef __JITI_H__
#define __JITI_H__

#include <stdio.h>
#include <unistd.h>
#include <string.h>

#ifdef EBUG
# define DEBUG
#endif

#ifdef DEBUG
# define DBG(X) X
#else
# define DBG(X)   
#endif

#include "dis-asm.h"
#include <pthread.h>
#include <stdlib.h>

#define TRUE  1
#define FALSE 0

/* MEMTYPE */
#define LOAD   0
#define STORE  1
#define MODIFY 2

#define TID(X)   (X+1023)/1024
#define THREADID TID(pthread_self())  

#define max(a,b) ((a)>(b)?(a):(b))
#define min(a,b) ((a)<(b)?(a):(b))


// special callbacks
#define CB_EXIT (char*)1
#define CB_MEM  (char*)2
#define CB_CODE (char*)3
#define CB_INIT (char*)4

extern unsigned _instrument_code_;
extern unsigned _instrument_mem_;
extern unsigned _instrument_dyn_;

extern unsigned _optimize_;

extern unsigned tot_bp;
extern unsigned main_ret;
extern unsigned delta;
extern unsigned ddelta;
extern unsigned end;
extern unsigned exit_jmpaddr;
//extern unsigned addr;
//extern unsigned len;

extern FILE * jiti_tty;

extern unsigned ori_start, ori_stop, clone_start, clone_stop;

extern void __wait_and_set(unsigned);

extern int calc_leal_len(char * instr, unsigned len);
unsigned is_plt_entry(unsigned address);
extern void panic(char *);
extern char _jiti_naam[100];

//void iret_helper(void);

#define BP 0xCC   /* int 3 */

#define CALL_IMM   0xE8

// Volgende instructie flusht de reeds geprefetchte instructies.
// Dit volstaat voor self-modifying code
#define FLUSH_INSTRUCTIONS asm(".word 0x00EB")  // JMP +0

/////////////////////////////////////////////////////////

#define GET_CHAR(addr)                                  \
     (*(char*)(addr)&0xFF)

#define PUT_CHAR(addr, value)                           \
     *(char*)(addr)=(value)

#define PUT_SHORT(addr,value)                           \
     *(unsigned short*)(addr)=(value)

#define GET_SHORT(addr)                                 \
     *(unsigned short*)(addr)

#define GET_LONG(addr)                                  \
     *(unsigned*)(addr)

#define PUT_LONG(addr, value)                           \
     *(unsigned*)(addr)=(unsigned)(value)

#define PUT_COPY(addr, delta, len)                      \
     memcpy((char*)(addr), (char*)(addr)-delta, len)

/////////////////////////////////////////////////////////

#define PUT_CALL(addr, rel_target)                      \
     PUT_CHAR(addr, CALL_IMM);                          \
     PUT_LONG(addr+1, rel_target);

#define RESTART_INSTRUCTION                             \
     JUMP_TO(address); 

#define SKIP_INSTRUCTION           

/////////////////////////////////////////////////////////

#define SIZE(opcode)                                    \
     ((opcode)&1)

#define DIRECTION(opcode)                               \
     (((opcode)&2)>>1)

/////////////////////////////////////////////////////////

/* instrumenteer instructie */
#define INSTR_CALL(handler)                             \
     PUT_CALL(address, (unsigned)handler-address-5);    \
     RESTART_INSTRUCTION;

/////

#define PUSHL32(addr, imm32)                           \
     PUT_CHAR(addr, 0x68);                             \
     PUT_LONG(addr+1, imm32);                          


#define PUSHL8(addr, imm8)                             \
     PUT_CHAR(addr, 0x6A);                             \
     PUT_CHAR(addr+1, imm8);                           


// note: the order in which the following is done is important to avoid
// data races in LOOKUP_CLONE2INSTR() when instrumenting MT programs
#define JMPL(addr, disp32)                             \
     PUT_LONG(addr+1, disp32);                         \
     PUT_CHAR(addr, 0xE9);

#define MOVL_EAX_2_MEM(addr, memaddr)                  \
     PUT_CHAR(addr, 0xA3);                             \
     PUT_LONG(addr+1, memaddr);                        


#define MOVL_MEM_2_EAX(addr, memaddr)                  \
     PUT_CHAR(addr, 0xA1);                             \
     PUT_LONG(addr+1, memaddr);                        


/* addl $delta,(%esp) */
#define ADDL_ESP(addr, delta)                          \
     PUT_CHAR(addr, 0x81);                             \
     PUT_CHAR(addr+1, 0x04);                           \
     PUT_CHAR(addr+2, 0x24);                           \
     PUT_LONG(addr+3, delta);                          

/* addl $delta, 4(%esp) */
#define ADDL_4ESP(addr, delta)                         \
     PUT_LONG(addr, 0x04244481);                       \
     PUT_LONG(addr+4, delta);                          

#define PUSHF                                          \
     PUT_CHAR(X_instr, 0x9C);  

#define POPF                                           \
     PUT_CHAR(X_instr, 0x9D);  

#define RET                                            \
     PUT_CHAR(X_instr, 0xC3);  

#define PUSH_REG(reg)                                  \
     PUT_CHAR(X_instr, 0x50|reg);

#define POP_REG(reg)                                   \
     PUT_CHAR(X_instr, 0x58|reg);

#define EAX 0
#define EBX 3
#define ECX 1
#define EDX 2
#define ESP 4
#define EBP 5
#define ESI 6
#define EDI 7
#define GS  8

/////////////////////////////////////////////////////////

typedef struct _clone2instr_ {
  unsigned from;
  unsigned to;
} _clone2instr_t;

// power of two to make mod'ing fast
#define CLONE2INSTR_SIZE (1 << 20)

extern _clone2instr_t _clone2instr_[CLONE2INSTR_SIZE];


unsigned register_callback(char * function, unsigned callback);
unsigned decode_block(unsigned address);
void process_elf_info(char * executable);
extern unsigned get_bp();
void par_init();
void _jiti_init_addr2line(char * _jiti_naam);
void _jiti_init_info();
unsigned intercept_plt(char * name, unsigned new_target);
unsigned LOOKUP_RELOC(unsigned address);
unsigned no_mem(unsigned, unsigned, unsigned);
void panic(char * string);
unsigned check_mod(unsigned imm_len, unsigned len, unsigned type);
void skip();
void add_to_plt(char * name, unsigned plt_address);
unsigned LOOKUP_CLONE2INSTR(unsigned);
unsigned get_exit_jmpaddr(void);
int jiti_parallel_program();


char * jiti_program_name(void);

extern char * jiti_version;

/* Dit moet op het einde staan. Bovendien moet jiti.h als laatste
   headerfile geincludeerd worden!! */

#define INIT(backend_callback)\
 _init(){jiti_init(backend_callback);}

#define CALL_FUNCTION(XXX)\
  ((void(*)(void))(XXX))()

// voor de backends
void jiti_lock();
void jiti_unlock();

#endif //__JITI_H__
