/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  instrument.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: jmaebe $
 * $Date: 2002/03/26 13:15:52 $
 * $Log: instrument.c,v $
 * Revision 1.7.2.38  2002/03/26 13:15:52  jmaebe
 * * fixed detection of PLT jmp_mem entries
 * * if we detect that we are going to return to a breakpoint in
 *   LOOKUP_RET_ADDRESS, we immediately instrument that code
 * * the jump coming after instrumented "int 0x80"'s is now replcaced with a
 *   a call to jmpcall_dir_handler too
 * * the clone wasn't refilled partially with breakpoints when we followed a
 *   direct short jump (with instruction length 2)
 *  -> with the 3 previous changes, we now only touch breakpoints when starting
 *   the program and when returning from the first call to a routine in
 *   libpthread (-> FPC has only one breakpoint left, pdf2ps 7)
 *
 * Revision 1.7.2.37  2002/03/18 21:45:13  jmaebe
 * + support for patching absolute addresses
 * + patch return address of calls to libpthread routines so that we take less
 *   breakpoints when they are used
 * * when there is an instruction that can't be replicated in the clone
 *   (because it needs extra instrumention code) and we can't put a link
 *   between the clone and the instrumentation code (because the instruction
 *   is < 5 bytes), replace all instructions replicated in the clone between
 *   the previous link from the clone to the instrumented code and the
 *   current instruction with breakpoints. Previously, if we regularly returned
 *   in the clone on a place where there were some replicated instructions
 *   followed by a breakpoint, we would always take the breakpoint because no
 *   clone2instr entry was ever added for the replicated instruction address.
 * + not followed direct calls and jumps are now handled using a mechanism
 *   similar to that for conditional jumps: once they are followed, they are
 *   patched so they always go to instrumented code and never end up at a
 *   breakpoint
 * * fixed loop support
 * + dynlib calls that have a "pushl imm32" before their "jmp *mem" in their
 *   plt entry are now also optimized
 * * jcc_taken_handler() doesn't lookup anything anymore in the reloc info
 *   because that's not thread safe (and it (almost?) never found anyting in
 *   there anyway, so it isn't worth it to fix that)
 * * improved patch info a bit: all patches are now tried at least twice,
 *   since it's not very unlikely that the patch will succeed after the next
 *   block has been instrumented. This new code is a bit ugly and needs to
 *   be cleaned up/commented though.
 *
 * Revision 1.7.2.36  2002/03/15 15:41:04  jmaebe
 * + support for bsf (for ghostscript)
 * * fixed support for bsr (idem)
 * * jmp_mem now also generates a call to adjust_call_target_ instead of to
 *   adjust_jump_target_ because some libc routines jump via a mem location
 *   to code in libpthread
 * * fixed some trace-code issues with the dynlib procedure call optimizations
 *
 * Revision 1.7.2.35  2002/03/14 17:29:26  jmaebe
 * * optimized the hell out of the reloc/patch facilities -> on moore a gzip
 *   operation is only 20% slower with the opt backend than without!!
 *
 * Revision 1.7.2.34  2002/03/12 18:02:18  jmaebe
 * + "xaddl reg,mem" and "shldl cl, reg, reg" support (for Mozilla)
 * * check whether there's really a "jmp mem" at the start of plt entries
 *   before doing optimizations, that's not always the case (in Mozilla
 *   there's first a "push const32"). After these two fixes, Mozilla still
 *   segfaults though :(
 *
 * Revision 1.7.2.33  2002/03/12 09:01:36  jmaebe
 * * removed some warnings and an obsolete comment
 *
 * Revision 1.7.2.32  2002/03/11 21:21:59  jmaebe
 * * MAKE_LINK is now no longer used while generating code, since it also
 *   caused problems when using MT code (thread 1 is instrumenting, generats a
 *   link to its instrumented code which isn't finished yet, thread 2 returns
 *   from a function or jumps/calls to the linked address and starts executing
 *   the unfinished instrumented code) -> changed to SCHEDULE_MAKE_LINK() and
 *   committed after the current block is finished
 * * LOOKUP_CLONE2INSTR() once again uses the jumps in the clone to determine
 *   targets
 * * libpthread is no longer instrumented because it disables all signals
 *   (using syscall 0xb3) in some routines, causing JiTI to hang (since its
 *   breakpoints aren't processed anymore, it can't continue instrumenting and
 *   the program can't continue executing)
 * * made jcc_taken_handler() thread-safe: the MAKE_LINK() there was not safe
 *   since it consists of 2 PUT_****() operations, so another thread could be
 *   executing the original call while it was being replaced with a JMP. At
 *   the same time, PATCH_INFO was made obsolete by doing the patching inside
 *   jcc_taken_handler (and that should be more or less thread safe, since it
 *   replaces exactly one int) -> GEBRUIK_PATCH_INFO is always set to 0 in
 *   jiti.c, can possibly be removed.
 * + infrastructure for intercepting syscalls and modifying the values passed
 *   to the kernel, not used anymore (I used it to debug the syscall 0xb3
 *   problem)
 * * separate assembler handlers for adjusting the target of a jmp or call
 *   mem/reg when instrumenting parallel programs (the call handler checks
 *   for targets in libpthread, the jump handler doesn't). The common handler
 *   is still used if we aren't instrumenting a parallel program (and that
 *   one obviously doesn't include the pthread checks)
 * * the memory layout is now always written to the log
 * - commented out some unused global variables in instrument.c
 * * some cosmetic changes and translations
 * * with the above changes, ploop now always works (on a dual proc machine)
 *   with the opt and trace-mem backends). ptest now always crashes for some
 *   reason however. If you use trace-code, you also almost always get crashes
 *   with MT programs.
 * And that's all for today, folks!
 *
 * Revision 1.7.2.31  2002/03/08 21:56:45  jmaebe
 * - disabled use of jumps in clone LOOKUP_CLONE2INSTR because it's not
 *   thread-safe (a running thread could get one of those targets as result
 *   while another thread is still creating the X_instr block this target
 *   leads to). Possible solution: create a SCHEDULE_MAKE_LINK() like the
 *   already introduced SCHEDULE_UPDATE_CLONE2INSTR() and only finish the
 *   links after a whole block has been instrumented.
 *
 * Revision 1.7.2.30  2002/03/08 21:30:02  jmaebe
 * * made jiti_fprint_ins1 thead-safe (the disassembler isn't re-entrant)
 *
 * Revision 1.7.2.29  2002/03/08 20:12:05  jmaebe
 * + SCHEDULE_UPDATE_CLONE2INSTR() and corresponding COMMIT_CLONE2INSTR()
 *   methods which allow addresses in the middle of an instrumented sequence
 *   to be added thread-safely to the clone2instr hashtable
 * + optimize calling of dynlib routines via plt (NOTE: requires that
 *   LD_BIND_NOW=YES)
 *
 * Revision 1.7.2.28  2002/03/08 16:13:33  jmaebe
 * + optimize dynlib function preludes ("call .+5; popl %ebx; addl $xxx,%ebx"
 *   and "call helper; addl $xxx,%ebx" with helper = "movl (%esp),%ebx; ret")
 *   changed to "movl $.+5+xxx, %ebx)
 * * fixed bug in call_dir and jmp_dir handlers where in some cases a wrong
 *   link from the clone to X_instr was created
 * - remove X_len global variable
 *
 * Revision 1.7.2.27  2002/03/07 17:26:22  jmaebe
 * * jmp/call mem/reg targets are now also looked up in clone2instr etc (using
 *   the LOOKUP_RET_HANDLER())
 * * helpers foor looking up targets and jcc_taken_handler in handlers.s now
 *   only save registers that are not saved by the standard calling conventions
 *   (eax, ecx and edx)
 * * moved code to add a target to the clone2instr table to a separate procedure
 * * increased clone2instr cache size to 4MB (1M entries)
 * * when following direct calls, the entry points of those procedures are
 *   added now to clone2instr as well. Doing the same for jumps seemed to only
 *   slow down everything.
 * * total speed improvement (for FPC): 27%
 * * decrease in size of generated instrumented code (for FPC): 50%
 *
 * Revision 1.7.2.26  2002/03/07 16:18:12  jmaebe
 * + remove jumps at the end of instrumented code blocks if they are to the
 *   clone address corresponding to the address of the instruction we start
 *   to instrument in a next block
 * * some formatting cleanups
 *
 * Revision 1.7.2.25  2002/03/06 20:54:55  jmaebe
 * * sped up jiti by about 1000% by using a hashing function to lookup entries
 *   in clone2instr. FPC now runs only about 4 times slower with instrumentation
 *   than without.
 *
 * Revision 1.7.2.24  2002/03/06 12:49:23  jmaebe
 * * improved LOOKUP_CLONE2INSTR significantly
 * * corrected behaviour of FOLLOW_JUMP/CALL_DIR (we followed jumps/calls if
 *   those vars were false instead of true
 * * adapted the JMPL macro so that LOOKUP_CLONE2INSTR remains thread-safe
 *
 * Revision 1.7.2.23  2002/03/06 11:03:31  jmaebe
 * * some cleanups
 * * always use clone2instr in call_dir_handler, also when USE_RELOC is false
 *
 * Revision 1.7.2.22  2002/03/02 21:20:46  jmaebe
 * * fixed very annoying bug in my patch so that breakpoints for "call mem"'s
 *   with len >= 5 in the clone were replaced by a jmp to the corresponding
 *   code in X_instr: the jmp's were being placed by the jmp_mem handler, so
 *   the jmp endeed up after the push of the return address
 * * fixed looking up of targets in the CLONE2INSTR by the PATCH stuff
 *   (delta wasn't added to the looked up targets-> never produced a hit)
 *
 * Revision 1.7.2.21  2002/03/01 09:46:31  jmaebe
 * * in some cases a wrong from-address was stored in _clone2instr_ after my
 *   previous changes
 *
 * Revision 1.7.2.20  2002/03/01 08:43:19  jmaebe
 * * fixed bug in mem instrumentation of btx instructions
 *
 * Revision 1.7.2.19  2002/02/28 15:53:34  jmaebe
 * * signal handlers are re-entrant under Linux nowadays, but bp_handlet set/
 *   used a global variable, causing general havoc with multi-threaded programs
 *   -> bp changed to a local variable (it isn't used outside bp_handler anyway)
 * * __test_and_set() prototype moved from insturment.c to jiti.h
 * * clone2instr updating now happens after the current block has been finished,
 *   otherwise a thread could start executing the unfinished block of
 *   instructions that another thread is generating
 * * interrupt_handler() only sets X_kloon to 1 if we aren't instrumenting code
 * * removed unnecessary locks in jcc_taken_handler()
 *
 * !!WARNING: don't ever use pthread_mutex_lock() etc, because these functions
 *   don't seem to be re-entrant so we hang if we call them while we are
 *   instrumenting them.
 *
 * Revision 1.7.2.18  2002/02/27 17:40:29  jmaebe
 * * fixed some more entries in the opcode0F_table
 *
 * Revision 1.7.2.17  2002/02/27 15:03:14  jmaebe
 * * made RELOC and CLONE2INSTR thread-safe
 * * fixed use of UPDATE_RELOC macro
 * * moved LOOKUP_CLONE2INSTR from jiti.c to instrument.c
 * * fixed usage of LOOKUP_CLONE2INSTR (in most cases, an address in the
 *   original program instead of in the clone was passed, so it never found
 *   anything)
 *
 * Revision 1.7.2.16  2002/02/27 11:34:41  jmaebe
 * * fixed data races when instrumenting MT programs
 *
 * Revision 1.7.2.15  2002/02/27 10:19:25  jmaebe
 * * check if LOOKUP_RELOC() returns a value != 0 in loop() (in case someone
 *   uses loop in an unorthodox way and "loops" to an address ahead)
 * * decreased the INSTR_LEN's passed to no_mem in the opcode_0F table by 1
 *   since the 0F prefix is already counted in the ORIGINAL_INSTRUCTION now
 *   (this fixes a lot of problems!!)
 * + when using DEBUG, also add the address of the place where the return
 *   address from main is stored on the stack
 *
 * Revision 1.7.2.14  2002/02/25 15:29:23  jmaebe
 * * fixed some DEBUG code (activating DEBUG shoudln't crash jiti anymore)
 * + breakpoints corresponding to jmp/call_mem's in the clone are now
 *   replaced with jumps to X_instr-code if their length is >= 5 (this is the
 *   case for e.g. the jmp_mem's in the plt)
 *
 * Revision 1.7.2.13  2002/02/25 14:26:57  jmaebe
 * + jecxz support
 *
 * Revision 1.7.2.12  2002/02/22 17:03:42  jmaebe
 * * made checking of ret from main thread safe
 * * LOOKUP_RET_HANDLER now also uses LOOKUP_CLONE2INSTR and LOOKUP_RELOC when
 *   the ret address is in a dynlib
 * * ret_handler (handlers.s) now saves all registers (it calls the C-routine
 *   LOOKUP_RET_HANDLER), should be slightly faster and is now jumped at
 *   instead of called
 * * fac now works reliably too when instrumented!!
 *
 * Revision 1.7.2.11  2002/02/22 15:42:49  jmaebe
 * * fixed checking of ret from main. tt and "hello world" now work!!!!!!!
 *
 * Revision 1.7.2.10  2002/02/22 14:48:34  jmaebe
 * * fixed gs override support
 *
 * Revision 1.7.2.9  2002/02/22 14:23:31  jmaebe
 * + save return address of main at the start of the program and use it to
 *   stop instrumenting when we return from main
 *
 * Revision 1.7.2.8  2002/02/21 13:49:12  jmaebe
 * * both the loading of the value and the storing to the stack is now
 *   instrumented for "push mem" (merged)
 * + added prototype for get_mem_len
 *
 * Revision 1.7.2.7  2002/02/21 13:03:10  ronsse
 * Added support for CMOVcc instructions.
 *
 * Revision 1.7.2.6  2002/02/20 12:24:51  jmaebe
 * - removed all "ifdef INSTRUMENT_DYNLIBS" directives (always enabled now)
 * * merged changes from r1.13-r1.19 from main branch
 * - removed no_mem_prefixed handler (not needed anymore because of simplified
 *   prefix handling)
 *
 * Revision 1.7.2.5  2002/02/15 15:46:48  jmaebe
 * * merged changes from revisions 1.11 and 1.12 from main branch
 *   (call_mem_handler fix, "ret imm" support)
 *
 * Revision 1.7.2.4  2002/02/14 17:47:32  jmaebe
 * + support for "lock" (merged, including fix)
 * + support for cmpxchg (merged)
 * * made jmp/call reg/mem again even slightly faster :)
 * + support for in/out
 *
 * Revision 1.7.2.3  2002/02/13 15:48:25  jmaebe
 * * optmized jmp_mem/call_mem and jmp_reg/call_reg handling
 * * fixed interrupt_handler()
 *
 * Revision 1.7.2.2  2002/02/13 12:16:58  jmaebe
 * + support for interrupts
 * * instrumenting dynamic libraries now works, but libc still causes some
 *   problems because of unsupported instructions (e.g lock and cmpxchgl)
 *
 * Revision 1.7.2.1  2002/02/12 16:47:20  jmaebe
 * + initial implementation of instrumenting of dynamically loaded libraries,
 *   but not yet working (even main(void) {} crashes :). Also, some backends
 *   which have to be modified before they can work with this new feature
 *   haven't been adapted yet. Compiler with -DINSTRUMENT_DYNLIBS to include
 *   the functionality (it's even required currently, because there's a
 *   reference to ddelta (dynlib delta) in handlers.s which can't be disabled
 *   using a conditional. Method used: separate clone for the dynlibs (thanks
 *   for the idea, Michiel :)
 *
 * Revision 1.7  2002/02/07 13:19:52  jmaebe
 * * replaced occurrences of "PUT_CHAR(X_instr, 0x9C/0x9D)" with PUSHF/POPF
 * * slightly optimized jmp_mem_handler:
 *   * changed "pushf;addl $delta,%eax; popf" to "leal delta(%eax),%eax"
 *   * changed "pushl $0" (used to reserve room on the stack for the new
 *     target address) to "pushl %eax" (smaller machine code)
 *
 * Revision 1.6  2002/01/10 21:23:33  ronsse
 * *** empty log message ***
 *
 * Revision 1.5  2001/11/12 10:08:39  ronsse
 * *** empty log message ***
 *
 * Revision 1.4  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


//#define DEBUG

#include "jiti.h"

#include "dynlib.h"

//////////////////////////////////////////////

// dummy argument voor arg1, arg2 en arg3
#define DUMMY 0

////// lock om de (parallelle) instrumentatie te beveiligen

//static pthread_mutex_t mmuutteexx;
//#define LOCK_DECODER()   pthread_mutex_lock(&mmuutteexx)
//#define UNLOCK_DECODER() pthread_mutex_unlock(&mmuutteexx)

volatile static int lloocckk = 0;
#define LOCK_DECODER()   DBG(fprintf(jiti_tty,"thread %d trying to get lock",THREADID);)__wait_and_set((unsigned)&lloocckk) DBG(;fprintf(jiti_tty,"thread %d got lock",THREADID))
#define UNLOCK_DECODER() DBG(fprintf(jiti_tty,"thread %d releasing lock",THREADID);)lloocckk=0

void jiti_lock()   { LOCK_DECODER();   }
void jiti_unlock() { UNLOCK_DECODER(); }


//////

#define INSTR_LEN(X) X
#define IMM_LEN(X)   X
#define MEM_LEN(X)   X
#define CALC         0
#define TYPE(X)      X

extern int _jiti_par_exec;

// optimalisaties
unsigned FOLLOW_JMP_DIR = 1;
unsigned FOLLOW_CALL_DIR = 1;
unsigned MAX_FOLLOW = 10;

unsigned GEBRUIK_RELOC_INFO = 1;
// power of 2 to make mod faster
#define RELOC 256

unsigned GEBRUIK_PATCH_INFO = 1;
// power of 2 to make mod faster
#define PATCH 128

// Note: when INSTRUMENT_DYNLIBS is defined, this *must* be 1! (JM)
unsigned LOOKUP_RET = 1;

// instrumentaties
unsigned _instrument_code_ = 0;
unsigned _instrument_mem_ = 0;
unsigned _instrument_dyn_ = 0;
//

unsigned _X_start = 0;
// if the previous instrumented block ended with a jump, this variable
// contains that jump's address
unsigned X_instr_end_jmp;
unsigned x_instr_size = 0;

extern void ret_handler_(void);
extern void jmpcall_dir_handler_(void);
extern void jcc_taken_handler_(void);
extern void adjust_jump_target_(void);
extern void adjust_call_target_(void);
extern void adjust_target_no_pthread_(void);
unsigned get_mem_len(unsigned byte2__NOT_USED);

unsigned follow = 0;

extern void trace_(void);
extern void trace_code_(void);

unsigned X_instr_start_block;

//// RELOCATIE INFO
volatile static int reloc_lock = 0;
#define LOCK_RELOC()   __wait_and_set((unsigned)&reloc_lock)
#define UNLOCK_RELOC() reloc_lock=0

unsigned reloc_start_i = 0, reloc_i = 0;
struct _reloc_info_ {
	unsigned _from_;
	unsigned _to_;
} _reloc_info_[RELOC];

#define USE_HASH
#ifdef USE_HASH
#	define HASH(key) (key%RELOC)
#	define UPDATE_RELOC(from, to) \
	if (GEBRUIK_RELOC_INFO) { \
		LOCK_RELOC(); \
		_reloc_info_[HASH((unsigned)from)]._from_ = from; \
		_reloc_info_[HASH((unsigned)from)]._to_   = to; \
		UNLOCK_RELOC(); \
	}
#else
#	define UPDATE_RELOC(from, to) \
	if (GEBRUIK_RELOC_INFO) { \
		LOCK_RELOC(); \
		_reloc_info_[reloc_i]._from_ = (unsigned)from; \
		_reloc_info_[reloc_i]._to_   = (unsigned)to;   \
		reloc_i=(reloc_i+1)%(RELOC); \
		if (reloc_i == (RELOC) / 2) \
			reloc_start_i = 0; \
		else if (reloc_i == 0) \
			reloc_start_i = (RELOC) / 2; \
		UNLOCK_RELOC(); \
	}
#endif

//// PATCH INFO
unsigned to_be_patched_start_i = 0, to_be_patched_i = 0;

struct _to_be_patched_info_ {
	unsigned _address_;	// adress where we have to patch
	unsigned _target_;	// target that has to be patched
	unsigned _absolute_;	// is the target absolute or relative
} _to_be_patched_info_[PATCH];

#define TO_BE_PATCHED(address, target,absolute) \
	if (GEBRUIK_PATCH_INFO) { \
		_to_be_patched_info_[to_be_patched_i]._address_ = (unsigned)address; \
		_to_be_patched_info_[to_be_patched_i]._target_  = (unsigned)target;   \
		_to_be_patched_info_[to_be_patched_i]._absolute_  = (unsigned)absolute;   \
		to_be_patched_i=(to_be_patched_i+1)%PATCH;\
		DBG(fprintf(jiti_tty, "\n>>>[TO BE: <%x : %x>]", address, target));\
	}

unsigned X_address;	/* algemene info */
unsigned X_kloon;		/* is het ok om de instr in de kloon uit te voeren? */

//char X_instr_[64*1024*1024];

char *X_instr_ = 0;		// start v/d aangemaakte code
char *X_instr = 0;		// plaats waar nieuwe code geplaatst wordt
static unsigned last_clone_link = 0;	// last place where we had a link from the clone to the X_instr

volatile static int clone2instr_lock = 0;
#define LOCK_CLONE2INSTR()   __wait_and_set((unsigned)&clone2instr_lock)
#define UNLOCK_CLONE2INSTR() clone2instr_lock=0
#define CLONE2INSTR_HASH(X) ((X) % CLONE2INSTR_SIZE)
#define CLONE2INSTR_SCHEDQUEUE 15
void UPDATE_CLONE2INSTR(unsigned clone_address, unsigned X_instr_address);
void SCHEDULE_UPDATE_CLONE2INSTR(unsigned clone_address, unsigned X_instr_address);
void COMMIT_CLONE2INSTR(void);
_clone2instr_t _clone2instr_[CLONE2INSTR_SIZE];
static _clone2instr_t _scheduled_clone2instr_[CLONE2INSTR_SCHEDQUEUE];
static unsigned _clone2instr_sched_count = 0;

#define MAKE_LINK_SCHEDQUEUE 1000
void SCHEDULE_MAKE_LINK(unsigned from, unsigned to);
void COMMIT_MAKE_LINK(void);
static unsigned _scheduled_make_link_[MAKE_LINK_SCHEDQUEUE];
static unsigned _make_link_sched_count = 0;

#define TRACE_CODE(from, to) \
	if (_instrument_code_) { \
		DBG(fprintf(jiti_tty, "\n>>>TRACE_CODE(%x, %x)", from, to)); \
		PUSHL32(X_instr, from);    X_instr+=5; \
		PUSHL32(X_instr, to);      X_instr+=5; \
		PUT_CALL(X_instr, trace_code_-(unsigned)X_instr-5);  X_instr+=5; \
	}

unsigned _fp(unsigned dummy1, unsigned dummy2, unsigned dummy3);
unsigned _mmx(unsigned dummy1, unsigned dummy2, unsigned dummy3);
unsigned _3dnow(unsigned dummy1, unsigned dummy2, unsigned dummy3);
unsigned _sse(unsigned dummy1, unsigned dummy2, unsigned dummy3);

//unsigned char opcode1;
//unsigned from_address;
//unsigned to_address;

unsigned extended0F(unsigned, unsigned, unsigned);

//unsigned len;
unsigned fetch_address;

#include <stdio.h>
extern FILE *jiti_tty;

unsigned char byte1, byte2, byte3, byte4, prefix_c, immsize_override, lock, seg_override;

/////////////////////

#define GIVE_INFO\
	jiti_fprint_ins1(jiti_tty, 1, fetch_address);\
	fprintf(jiti_tty, "\n%x %x %x %x\n", byte1, byte2, byte3, byte4)

unsigned todo(unsigned dummy1, unsigned dummy2, unsigned dummy3){
  	GIVE_INFO;
	panic("TODO: This instruction is not yet instrumented!");
	return 0;
}

unsigned dontexist(unsigned dummy1, unsigned dummy2, unsigned dummy3){
  	GIVE_INFO;
	panic("TODO: This instruction doesn't exist. Go away!");
	return 0;
}

unsigned unsupported(unsigned dummy1, unsigned dummy2, unsigned dummy3){
  	GIVE_INFO;
	panic("TODO: This instruction isn't supported. Go away!");
	return 0;
}

unsigned privileged(unsigned dummy1, unsigned dummy2, unsigned dummy3){
  	GIVE_INFO;
	panic("TODO: This is a privileged instruction!.");
	return 0;
}

///////////////////////////////////////////////////////

void STACK_ACCESS(unsigned type, unsigned size){

	if (lock)
		panic("lock set in STACK_ACCESS??");

	DBG(fprintf(jiti_tty, "\n>>> STACK_ACCESS(%d)", type));
	if (_instrument_mem_) {
		// stack_access is called for push/pop/pusha/popa. When pushing (type ==
		// TYPE(STORE)), the value(s) is (are) stored between ESP-size and ESP
		// -> we need to pass ESP-size to trace! By adjusting ESP now (instead of
		// substracting size from the pushed ESP later), we can guarantee that
		// the callback can also access this memory in case it wants to
		if (type == TYPE(STORE)) {
			if (size > 127)
				panic("size > 127 in STACK_ACCESS!!");

			PUT_LONG(X_instr,(-(signed char)size) << 24 | 0x24648d);
			X_instr += 4;		// leal -size(%esp),%esp
		}
		
		PUSH_REG(ESP);               X_instr++;       // pushl %esp
		PUSHL8(X_instr, size);       X_instr += 2;		// pushl len
		PUSHL8(X_instr, type);       X_instr += 2;		// LOAD / STORE
		PUSHL32(X_instr, X_address); X_instr += 5;		// pushl X_address
		PUT_CALL(X_instr, trace_ - (unsigned) X_instr - 5);
		X_instr += 5;
	
		if (type == TYPE(STORE)) {
			// restore ESP
			PUT_LONG(X_instr,size << 24 | 0x24648d);
			X_instr += 4;		// leal size(%esp),%esp
		}
	} else
		X_kloon = 1;
}

///////////////////

#define MAKE_LINK(from, to) \
	DBG(fprintf(jiti_tty,"\n>>>MAKE_LINK(%x, %x)", (from), (to)-(from)-5));\
	JMPL((from), (to)-(from)-5);

///////////////////

void CLEAN_CLONE_UNTIL(unsigned clone_address) {
	// remove all instructions that were copied to the clone until now
	DBG(fprintf(jiti_tty,"Cleaning from 0x%x till 0x%x\n",last_clone_link,clone_address));
	memset((char *)last_clone_link,BP,clone_address-last_clone_link);
	last_clone_link = clone_address;
}

///////////////////

void ORIGINAL_INSTRUCTION(unsigned instr_len){
	
	instr_len += prefix_c;
	// clear lock (it's counted with prefix_c)
	lock = 0;

	if (instr_len >= 5) {
		//    fprintf(jiti_tty, "LEN>=5, X_instr_start_block=%x\n", X_instr_start_block); 
		if (X_address > end) {
			SCHEDULE_MAKE_LINK(X_address + ddelta, X_instr_start_block);
		}
		else {
			SCHEDULE_MAKE_LINK(X_address + delta, X_instr_start_block);
		}
	} else {
		if (X_kloon) {
			if (X_address > end)
				memcpy((void *) (X_address + ddelta), (void *) X_address, instr_len);
			else
				memcpy((void *) (X_address + delta), (void *) X_address, instr_len);	// else { houdt BP; }
		}
	}

	memcpy((void *) X_instr, (void *) X_address, instr_len);
	X_instr += instr_len;
	X_address += instr_len;
}

////////////////////

unsigned leave() {
	if (lock)
		panic("lock before leave??");
	CLEAN_CLONE_UNTIL(X_address + ((X_address > end) ? ddelta: delta));
	PUT_CHAR(X_instr, 0x89);
	X_instr++;			/* movl %ebp, %esp */
	PUT_CHAR(X_instr, 0xEC);
	X_instr++;
	STACK_ACCESS(TYPE(LOAD),4);
	POP_REG(EBP);
	X_instr++;			/* popl %ebp */
	return X_address + 1;
}

////////////////////

unsigned nop(unsigned instr_len, unsigned dummy1, unsigned dummy2) {

	if (lock)
		panic("lock before nop??");

	if (X_address > end)
		*(unsigned char *) (X_address + ddelta) = 0x90;	/* nop plaatsen in kloon, en niets in instr */
	else
		*(unsigned char *) (X_address + delta) = 0x90;	/* nop plaatsen in kloon, en niets in instr */
	return X_address + instr_len;
}

////////////////////

unsigned movl_eax_mem_handler(unsigned instr_len, unsigned mem_len, unsigned mem_type){
	if (_instrument_mem_) {
		PUSHL32(X_instr, GET_LONG(X_address + 1 + prefix_c));
		X_instr += 5;
		PUSHL8(X_instr, mem_len);
		X_instr += 2;
		PUSHL8(X_instr, mem_type);
		X_instr += 2;
		PUSHL32(X_instr, X_address);
		X_instr += 5;
		PUT_CALL(X_instr, trace_ - (unsigned) X_instr - 5);
		X_instr += 5;
	}
	ORIGINAL_INSTRUCTION(instr_len);
	return X_address;
}

////////////////////////////////////////////////////////////


unsigned cmovcc(unsigned dummy1, unsigned dummy2, unsigned dummy3){


  if (byte2>>6==3) return no_mem(INSTR_LEN(2),IMM_LEN(0), 0);

  // Transformation:
  //       CMOVcc mem, reg
  // becomes
  //       Jcc X
  //       JMP Y
  //    X: code generated by check_mod()
  //    Y: CMOVcc mem, reg

  if (_instrument_mem_){
    PUT_CHAR(X_instr, byte1 | 0x30); X_instr++;             // Jcc
    PUT_CHAR(X_instr, 2); X_instr++;                        // disp8
    PUT_CHAR(X_instr, 0xEB); X_instr++;                     // jmp
    PUT_CHAR(X_instr, 19+get_mem_len(byte2)); X_instr++;    // disp8
  }

  return check_mod(IMM_LEN(0), MEM_LEN(4), TYPE(LOAD));

}

////////////////////////////////////////////////////////////

unsigned push_pop_handler(unsigned instr_len, unsigned type, unsigned dummy1){
	STACK_ACCESS(type,4);
	ORIGINAL_INSTRUCTION(instr_len);
	return X_address;
}

////////////////////////////////////////////////////////////////

unsigned push_pop_mem_handler(unsigned stacktype, unsigned memtype, unsigned dummy3){

	STACK_ACCESS(TYPE(stacktype),4);
	X_kloon = 0;
	return check_mod(IMM_LEN(0),MEM_LEN(4),TYPE(memtype));
}

////////////////////////////////////////////////////////////////

void expl_handler(unsigned leal_len, unsigned mem_len, unsigned mem_type){

	/* bewaar EAX */
 
	// we moeten een lokale variabele gebruiken voor par. progr.
	PUSH_REG(EAX);
	X_instr++;			// pushl %eax

	// hier moeten we nog testen of de adresexpressie SP gebruikt!!!! Indien ja, moet de
	// berekende waarde gewijzigd worden!!

	/* maak leal ..., %eax */
	memcpy((void *) X_instr, (void *) (X_address + prefix_c), leal_len);	/* kopieer */
	PUT_CHAR(X_instr, 0x8D);	/* lea */
	PUT_CHAR(X_instr + 1, GET_CHAR(X_instr + 1) & 0xC7);	/* , %eax */
	X_instr += leal_len;

	/* call trace_ */
	PUSH_REG(EAX);
	X_instr++;			/* pushl %eax */
	PUSHL8(X_instr, mem_len);
	X_instr += 2;
	PUSHL8(X_instr, mem_type);
	X_instr += 2;
	PUSHL32(X_instr, X_address);
	X_instr += 5;
	PUT_CALL(X_instr, trace_ - (unsigned) X_instr - 5);
	X_instr += 5;

	/* herstel EAX */
	POP_REG(EAX);
	X_instr++;			// popl %ax

	// MERK OP: geen originele instructie! 
}

////////////////////////////////////////////////////////////////////////////

unsigned get_mem_len(unsigned byte2__NOT_USED){
	unsigned char mod = byte2 >> 6;
	unsigned char rm = byte2 & 7;
	unsigned char base = byte3 & 7;

	//  printf("-B2=%x M%d RM%d B%d-", byte2, mod, rm, base);

	/* mod mag niet 3 zijn */
	if (mod == 3) panic("get_mem_len(reg1,reg2)");

	if ((mod == 0) && (rm != 4) && (rm != 5)) return 0;
	if ((mod == 0) && (rm == 5)) return 4;
	if ((mod == 1) && (rm != 4)) return 1;
	if ((mod == 2) && (rm != 4)) return 4;

	/* nu moet rm==4 zijn! => SIB */
	if (rm != 4) panic("get_mem_len: rm==4");

	if ((mod == 0) && (base != 5)) return 1;
	if ((mod == 0) && (base == 5)) return 5;
	if (mod == 1) return 2;
	if (mod == 2) return 5;		/* 5?? */

	/* hier mogen we nooit komen */
	panic("get_mem_len: not reached");
	return 0;
}

////////////////////////////////////////////////////////////////////////////

unsigned loop(unsigned dummy0, unsigned dummy1, unsigned dummy2){
	// loop* is 2-bytes lang en heeft een 8-bit displacement
	signed char disp8;
	unsigned target, new_target;

	if (lock)
		panic("lock before loop??");

	CLEAN_CLONE_UNTIL(X_address + ((X_address > end) ? ddelta: delta));

	disp8 = GET_CHAR(X_address + 1);
	target = X_address + (signed) disp8 + 2;

	// try to find a better target
	if ((new_target = LOOKUP_RELOC(target)) != 0)
		target = new_target;
	else if ((new_target = LOOKUP_CLONE2INSTR(target+(target>end)?ddelta:delta)) != 0)
		target = new_target;
	else
		target += (target > end) ? ddelta:delta;

	// Change
	//    LOOP X
	// to
	//    LOOP Y   ; 2 bytes
	//    JMP Z    ; 5 bytes
	// Y: JMP X    ; 5 bytes
	// Z:

	memcpy(X_instr, (char *) X_address, 1);           X_instr++;
	PUT_CHAR(X_instr, 5);                             X_instr++;
	JMPL(X_instr, 5);                                 X_instr += 5;
	JMPL(X_instr, target - (unsigned) X_instr - 5);   X_instr += 5;
	return X_address + 2;

}

////////////////////////////////////////////////////////////////////////////

unsigned no_mem(unsigned length, unsigned imm_len, unsigned dummy2){
	X_kloon = 1;

	if (length == CALC) {	/* we moeten zelf nog de lengte bepalen, enkel gebruikt voor LEAL */
		length = 2 + get_mem_len(byte2);
	} else {
		if (imm_len > immsize_override) {
			length -= immsize_override;
		}
	}

	ORIGINAL_INSTRUCTION(length);

	return X_address;

}

////////////////////////////////////////////////////////////////////////////

unsigned check_mod(unsigned imm_len, unsigned mem_len, unsigned type){
	unsigned leal_len, instr_len;
	unsigned char mod2 = byte2 >> 6;

	if (mod2 != 3) {		/* memory operation */

		leal_len = get_mem_len(byte2);

		if (imm_len) {
			instr_len = leal_len + imm_len;
			if (imm_len > immsize_override)
				instr_len -= immsize_override;	// 4 wordt 2
		} else {
			instr_len = leal_len;
		}

		leal_len += 2;
		instr_len += 2;

		//printf("@0x%x [IMM%d INSTRLEN%d LEALLEN%d OVERRIDE%d MEMLEN%d]\n", X_address, imm_len, instr_len, leal_len, immsize_override, mem_len); ////

		if (_instrument_mem_) {
			expl_handler(leal_len, mem_len, type);
			CLEAN_CLONE_UNTIL(X_address + ((X_address > end) ? ddelta: delta));
		}
		else
			X_kloon = 1;

		ORIGINAL_INSTRUCTION(instr_len);
		return X_address;

	} else {			/* reg,reg operation */
		return no_mem(INSTR_LEN(2 + imm_len), imm_len, DUMMY);
	}

}

//////////////////////////////////////////////////////////

unsigned check_mod_op2a(unsigned imm_len, unsigned mem_len, unsigned dummy){
	/* op2 testen voor L/X */
	unsigned type = 0;
	unsigned opcode2 = (byte2 & 0x3F) >> 3;

	switch (opcode2) {
		case 0:			/* add */
			type = TYPE(MODIFY);
			break;
		case 1:			/* or */
			type = TYPE(MODIFY);
			break;
		case 2:			/* adc */
			type = TYPE(MODIFY);
			break;
		case 3:			/* sbb */
			type = TYPE(MODIFY);
			break;
		case 4:			/* and */
			type = TYPE(MODIFY);
			break;
		case 5:			/* sub */
			type = TYPE(MODIFY);
			break;
		case 6:			/* xor */
			type = TYPE(MODIFY);
			break;
		case 7:			/* cmp */
			type = TYPE(LOAD);
			break;
		default:
			panic("check_mod_op2a: check type");
	}

	return check_mod(IMM_LEN(imm_len), MEM_LEN(mem_len), TYPE(type));
}

///////////////////////////////////////////////////////////

unsigned check_mod_op2b(unsigned imm_len, unsigned mem_len, unsigned type){
	/* op2 testen voor L/X */
	unsigned opcode2 = (byte2 & 0x3F) >> 3;

	switch (opcode2) {
		case 0:
			return check_mod(IMM_LEN(imm_len), MEM_LEN(mem_len), TYPE(LOAD));	// TEST
		case 2:
		case 3:
			return check_mod(IMM_LEN(0), MEM_LEN(mem_len), TYPE(MODIFY));	// NOT/NEG mem
		case 4:
		case 5:
		case 6:
		case 7:
			return check_mod(IMM_LEN(0), MEM_LEN(mem_len), TYPE(LOAD));	// MUL, IMUL, DIV, IDIV 
			// (E)A[XL] with mem
		default:
			printf("\n\n+++TODONOW: opcode2=%d+++\n\n", opcode2);
			_exit(-1);
	}
	// not reached
	return 0;

}

///////////////////////////////////////////////////////////

unsigned Poperand();
unsigned Plock();
unsigned PSegOverride(unsigned reg, unsigned dummy1, unsigned dummy2);

unsigned Paddress(){
	panic("--address size--");
	// not reached
	return 0;
}

#define REP_NONE  0
#define REP_LOAD  1
#define REP_STORE 2

extern unsigned range();

//////////////////////////////////////////////////////

void interrupt_modifier(unsigned eax, unsigned ebx, unsigned ecx, unsigned edx) {

	fprintf(jiti_tty,"syscall %u (0x%x), ebx = 0x%x, ecx = 0x%x, edx = 0x%x\n",eax,eax,ebx,ecx,edx);
	switch (eax) {
		case 179:
			fprintf(jiti_tty,"sigmask: size: %u, value of first 32bits: 0x%x\n",ecx,GET_LONG(ebx));
			break;
	}
}

//////////////////////////////////////////////////////

unsigned jmpcall_dir_handler(unsigned address) {
	unsigned target, new_target, instr_address, patch_address, target_delta;

	instr_address = GET_LONG(address);
	patch_address = address - 5 - 4;
	DBG(fprintf(jiti_tty,"In jmpcall_dir_handler, address = 0x%x, original instruction is at 0x%x and patch address is 0x%x\nOriginal instruction: ",address,instr_address,patch_address));
	DBG(jiti_fprint_ins1(jiti_tty,1,instr_address));

	switch GET_CHAR(instr_address) {
		case 0xE8: // call disp32
		case 0xE9: // jmp disp32
			target = GET_LONG(instr_address+1) + instr_address + 5;
			break;
		case 0xEB: // jmp disp8
			target = instr_address + ((signed char)GET_CHAR(instr_address+1)) + 2;
			break;
		case 0xCD: // int ($0x80)
			target = instr_address+2;
			break;
		default:
			printf("unknown instruction: 0x%.1x\n",GET_CHAR(instr_address)); 
			panic("Unknown call/jmp!!");
			// avoid warning
			target = 0;
			break;
	}

	DBG(fprintf(jiti_tty,"Target is 0x%x\n",target));

	target_delta = (target > end) ? ddelta:delta;

	// don't use reloc info, it can only be used inside the decoder
	// because it's not thread-safe.
	if ((new_target = LOOKUP_CLONE2INSTR(target+target_delta)) == 0) {
		new_target = decode_block(target+target_delta);
	}

	PUT_LONG(patch_address,new_target-patch_address-4);
	DBG(fprintf(jiti_tty,"jmpcall dir instruction after patch:"));
	DBG(jiti_fprint_ins1(jiti_tty,1,patch_address-1));
	FLUSH_INSTRUCTIONS;

	return new_target;
}


//////////////////////////////////////////////////////

extern void interrupt_modifier_(void);

unsigned interrupt_handler(unsigned dummy1, unsigned dummy2, unsigned dummy3){

	/* trace code */
	TRACE_CODE(_X_start, X_address+1);

/*
	PUSHL32(X_instr,X_address);
	X_instr+=5;
	PUT_CALL(X_instr,&interrupt_modifier_ - (unsigned)X_instr - 5);
	X_instr+=5;
*/
	if (!_instrument_code_)
		X_kloon = 1;
	else {
		CLEAN_CLONE_UNTIL(X_address + ((X_address > end) ? ddelta: delta));
	}

	/* replicate in X_instr (and in clone when not instrumenting code) */
	ORIGINAL_INSTRUCTION(2);

	/* jump to the next instruction */
	X_instr_end_jmp = (unsigned)X_instr;
	// create a dummy jump that will be patched later. Can't just
	// patch the call after this, because that would require two
	// operations -> not thread-safe
	JMPL(X_instr,0);
	X_instr+=5;
	PUT_CALL(X_instr,jmpcall_dir_handler_ - (unsigned) X_instr - 5);
	X_instr+=5;
	// ORIGINAL_INSTRUCTION already increased X_address with 2
	PUT_LONG(X_instr,X_address-2);
	X_instr+=4;
	/* don't instrument any further, this could be the end of the program */
	return 0;
}

//////////////////////////////////////////////////////

unsigned ret_handler(unsigned len, unsigned dummy2, unsigned dummy3){
	
	int retvalue;
	
	if (lock)
		panic("lock before ret??");

	CLEAN_CLONE_UNTIL(X_address + ((X_address > end) ? ddelta: delta));

	/* trace mem */
	STACK_ACCESS(TYPE(LOAD),4);

	/* trace code */
	TRACE_CODE(_X_start, X_address+len-1);

	if (len == 3) {
		// ret imm
		retvalue = GET_SHORT(X_address+1);
		
		// save eax
		PUSH_REG(EAX);
		X_instr++;

		// load the return address
		PUT_LONG(X_instr,0x0424448b);
		X_instr+=4;	// movl  0x4(%esp,1),%eax

		/* stack lay-out now:
		 * esp				: saved eax
		 * esp+4			: return address
		 * esp+8 till esp+8+retvalue	: arguments
		 */

		// save it at the start of the arguments
		if (retvalue+4 <= 127) {
			PUT_LONG(X_instr,(retvalue+4) << 24 | 0x244489);
			X_instr += 4;	// movl  %eax,imm8(%esp)
		} else {
			PUT_SHORT(X_instr,0x8489);
			PUT_CHAR(X_instr+2,0x24);
			PUT_LONG(X_instr+3,retvalue+4);
			X_instr += 7;	// movl %eax,imm32(%esp)
		}

		// restore eax
		POP_REG(EAX)
		X_instr++;
		
		// remove (old) return address and all arguments except for the first 4
		// bytes (where the return address is stored now)	
		if (retvalue <= 127) {
			// use short form
			PUT_LONG(X_instr,retvalue << 24 | 0x24648d);
			X_instr += 4;		// leal imm8(%esp),%esp
		} else {
			// use long form
			PUT_SHORT(X_instr,0xa48d);
			PUT_CHAR(X_instr+2,0x24);
			PUT_LONG(X_instr+3,retvalue);
			X_instr += 7;		// leal imm32(%esp),%esp
		}
	}

	JMPL(X_instr, ret_handler_ - (unsigned) X_instr - 5);
	X_instr += 5;

	return 0;			// einde van dit basisblok
}


//////////////////////////////////////////////////////

unsigned jmp_dir_handler(unsigned len, unsigned dummy1, unsigned dummy2){
	unsigned new_target, target, curdelta;
	signed char disp8;

	if (lock)
		panic("lock before jmp_dir??");

	if (len == 2) {		/* disp8  */
		disp8 = (signed) GET_CHAR(X_address + 1);
		target = disp8 + X_address + 2;
	} else {			/* disp32 */
		target = GET_LONG(X_address + 1) + X_address + 5;
	}

	// link from clone to here
	if (len >= 5) {
		if (X_address > end) {
			SCHEDULE_MAKE_LINK(X_address + ddelta, X_instr_start_block);
		} else {
			SCHEDULE_MAKE_LINK(X_address + delta, X_instr_start_block);
		}
	}

	// note: this delta is for the target, not for the X_address of the instruction
	// -> don't use above for MAKE_LINK() !!
	curdelta = (target > end) ? ddelta : delta;

	TRACE_CODE(_X_start, X_address + len - 1);

	if ((new_target = LOOKUP_CLONE2INSTR(target+curdelta))) {

		if (len < 5)
			CLEAN_CLONE_UNTIL(X_address + ((X_address > end) ? ddelta: delta));

		DBG(fprintf
				(jiti_tty,
				 "\n>>>Found target in clone2instr info! -> %x",
				 new_target));
		JMPL(X_instr, new_target - (unsigned) X_instr - 5);
		X_instr += 5;
		return 0;
	}

	if ((new_target = LOOKUP_RELOC(target))) {
		DBG(fprintf(jiti_tty, "\n>>>FOUND @ %x", new_target));

		if (len < 5)
			CLEAN_CLONE_UNTIL(X_address + ((X_address > end) ? ddelta: delta));

		//TRACE_CODE(_X_start, X_address+len-1); _X_start=target;
		JMPL(X_instr, new_target - (unsigned) X_instr - 5);
		X_instr += 5;
		return 0;
	}

	if (len < 5)
		CLEAN_CLONE_UNTIL(X_address + ((X_address > end) ? ddelta: delta));

	if (!FOLLOW_JMP_DIR || (follow++ > MAX_FOLLOW)) {
		X_instr_end_jmp = (unsigned)X_instr;

		// create a dummy jump that will be patched later. Can't just
		// patch the call after this, because that would require two
		// operations -> not thread-safe
		JMPL(X_instr,0);
		X_instr+=5;
		PUT_CALL(X_instr,jmpcall_dir_handler_ - (unsigned) X_instr - 5);
		X_instr+=5;
		PUT_LONG(X_instr,X_address);
		X_instr+=4;

		return 0;
	} else {
		DBG(fprintf(jiti_tty, "\n>>>Restart from %x", target));
		SCHEDULE_UPDATE_CLONE2INSTR(target+curdelta,(unsigned)X_instr);
		_X_start = target;
		last_clone_link = target + curdelta;
		DBG(fprintf(jiti_tty,"Following, last clone link = 0x%x\n",last_clone_link));
		return target;
	}

}

//////////////////////////////////////////////////////

unsigned jmp_reg_handler(unsigned is_jump, unsigned dummy2, unsigned dummy3){

	if (lock)
		panic("lock before jmp_reg??");

	CLEAN_CLONE_UNTIL(X_address + ((X_address > end) ? ddelta: delta));

	TRACE_CODE(_X_start, X_address + 1);

	PUT_CHAR(X_instr, (GET_CHAR(X_address + 1) & 7) | 0x50);
	X_instr++;			/* PUSHL %reg */
	// adjust target so it points to the clone and jump to it
	if (!jiti_parallel_program()) {
		JMPL(X_instr,((unsigned)&(adjust_target_no_pthread_))-(unsigned)X_instr-5);
	}
	else {
		if (is_jump) {
			JMPL(X_instr,((unsigned)&(adjust_jump_target_))-(unsigned)X_instr-5);
		}
		else {
			JMPL(X_instr,((unsigned)&(adjust_call_target_))-(unsigned)X_instr-5);
		}
	}
	X_instr+=5;
	return 0;
}

//////////////////////////////////////////////////////

unsigned jmp_mem_handler(unsigned len, unsigned is_jump, unsigned dummy2){
	if (lock)
		panic("lock before jmp_mem!!!");

	TRACE_CODE(_X_start, X_address + len - 1);

	if (len >= 5) {

		// link from clone to this code
		if (X_address > end) {
			SCHEDULE_MAKE_LINK(X_address + ddelta, X_instr_start_block);
		} else {
			SCHEDULE_MAKE_LINK(X_address + delta, X_instr_start_block);
		}
	}
	else {
		CLEAN_CLONE_UNTIL(X_address + ((X_address > end) ? ddelta: delta));
	}

	/* trace address */
	if (_instrument_mem_) {
		memcpy((void *) X_instr, (void *) X_address, len);	// copy original instruction
		PUT_CHAR(X_instr + 1, GET_CHAR(X_instr + 1) | 0x30 );	// turn the jmp/call into a push
		X_instr+=len;
		PUSHL8(X_instr, 4);
		X_instr += 2;
		PUSHL8(X_instr, TYPE(LOAD));
		X_instr += 2;
		PUSHL32(X_instr, X_address);
		X_instr += 5;
		PUT_CALL(X_instr, trace_ - (unsigned) X_instr - 5);
		X_instr += 5;
	}

	/* get target address */
	memcpy((void *) X_instr, (void *) X_address, len);	// copy original instruction
	PUT_CHAR(X_instr + 1, GET_CHAR(X_instr + 1) | 0x30 );	// turn the jmp/call into a push
	X_instr+=len;
	/* adjust target so it points to the clone and jump to it */
	if (!jiti_parallel_program()) {
		JMPL(X_instr,((unsigned)&(adjust_target_no_pthread_))-(unsigned)X_instr-5);
	}
	else {
		if (is_jump) {
			JMPL(X_instr,((unsigned)&(adjust_jump_target_))-(unsigned)X_instr-5);
		}
		else {
			JMPL(X_instr,((unsigned)&(adjust_call_target_))-(unsigned)X_instr-5);
		}
	}
	X_instr+=5;
	return 0;
}

//////////////////////////////////////////////////////

unsigned is_plt_entry(unsigned address) {
	unsigned opcode2;

	// sometimes there's first a push imm32
	if (GET_CHAR(address) == 0x68)
		address+=5;

	// check for "jmp mem"
	opcode2 = (GET_CHAR(address+1) & 0x3f) >> 3;
	if ((GET_CHAR(address) == 0xff) &&
	     ((opcode2 == 4) ||
	      (opcode2 == 5)) &&
	     ((GET_CHAR(address+1) >> 6) != 3)) {
		return TRUE;
	}
	return FALSE;
}

//////////////////////////////////////////////////////

unsigned call_dyn_handler(unsigned target) {

  unsigned real_target, new_target, real_target_delta;

	// a jump to a plt-entry, always arrives at a jmp_mem
	//  * if this entry hasn't been looked up yet, it contains the address of
	//   the routine in the dynlib (assume LD_BIND_NOW=YES)
	real_target = GET_LONG(GET_LONG(target+2));

	// Mozilla has PLT calls to its own code
	real_target_delta = (real_target > end) ? ddelta:delta;
	DBG(fprintf(jiti_tty,"target: 0x%x, real_target: 0x%x\n",target,real_target));
	if ((new_target = LOOKUP_CLONE2INSTR(real_target+real_target_delta)) != 0) {
		DBG(fprintf(jiti_tty, "dynlib routine at 0x%x found in X_instr at 0x%x\n", target, new_target));
		JMPL(X_instr, new_target-(unsigned)X_instr-5);
		X_instr+=5;
		// end of this block
		return 0;
	}
	else {
		DBG(fprintf(jiti_tty, "dynlib routine at 0x%x not yet instrumented or address not in cache anymore\n", target));
		// adjust jump target (instrumented code will follow now)

		// WARNING: we cannot adjust the entry in the PLT for the
		//   simple reason that if jiti also calls this function
		//   internally (e.g. fprintf for logging), it will then call the
		//   instrumented function too!!
		// PUT_LONG(GET_LONG(target+2),X_instr);
		// mark the start of the procedure in clone2instr
		SCHEDULE_UPDATE_CLONE2INSTR(real_target+real_target_delta,(unsigned)X_instr);
		last_clone_link = real_target+real_target_delta;
		_X_start = real_target;
		return real_target;
	}
}


//////////////////////////////////////////////////////

unsigned call_dir_handler(unsigned dummy1, unsigned dummy2, unsigned dummy3){

	unsigned new_target, curdelta;
	extern unsigned plt_end;
	unsigned target = GET_LONG(X_address + 1) + X_address + 5;

	if (lock)
		panic("lock before call_dir??");

	/* put link from clone to X_instr */
	if (X_address > end) {
		SCHEDULE_MAKE_LINK(X_address + ddelta, X_instr_start_block);
	} else {
		SCHEDULE_MAKE_LINK(X_address + delta, X_instr_start_block);
	}  

	// note: this delta is for the target, not for the X_address of the instruction
	// -> don't use above for MAKE_LINK() !!
	curdelta = (target > end) ? ddelta : delta;

	TRACE_CODE(_X_start, X_address + 4);

	if (_optimize_) {

		// optimize PIC prelude of dynlib functions

		// replace "call .+0; popl %ebx; addl $xxx, %ebx
		// by "movl $.+xxx,%ebx"
		// check if the call is followed by "popl %ebx; addl $xxx, %ebx
		if ((target == X_address+ 5) &&
				((GET_LONG(X_address+5) & 0x00ffffff) == 0x00c3815b)) {
			// replace
			PUT_CHAR(X_instr,0xbb);
			X_instr++; // movl $????,%ebx
			PUT_LONG(X_instr,X_address+5+GET_LONG(X_address+8));
			X_instr+=4; // ???? = curraddr+5+xxxx
			_X_start = X_address+5;
			return X_address+5+1+6;
		}

		//replace "call helper; addl $xxx, %ebx" with helper = "movl (%esp),%ebx; ret"
		// with the same as above
		// a) check for "movl (%esp),%ebx; ret" at the target
		if ((GET_LONG(target) == 0xc3241c8b) &&
		// b) check for "addl $xxx, %ebx"
				(GET_SHORT(X_address+5) == 0xc381)) {
			// replace
			PUT_CHAR(X_instr,0xbb);
			X_instr++; // movl $????,%ebx
			PUT_LONG(X_instr,X_address+5+GET_LONG(X_address+7));
			X_instr+=4; // ???? = curraddr+5+xxxx
			_X_start = X_address+5;
			return X_address+5+6;
		}
	}

	/* trace */
	STACK_ACCESS(TYPE(STORE),4);

	DBG(fprintf
			(jiti_tty, "\n>>>0x%x: CALL to 0x%x: %c PLT", X_address, target,
			 target < plt_end ? '<' : '>'));

	// returnadres wijst naar de ori. code, we wijzigen de bijhorende "ret" namelijk.
	PUSHL32(X_instr, X_address + 5);
	X_instr += 5;

	if ((target < plt_end) &&
	    is_plt_entry(target)) {
		unsigned new_ret_address, ret_address_address = (unsigned)X_instr - 4;

		// if the jump is to the exit entry, stop instrumenting
		if (target == (unsigned)&exit) {
			JMPL(X_instr,target-(unsigned)X_instr-5);
			X_instr+=5;
			return 0;
		}

		// some programs have a "pushl imm32" before the jmp_mem for some of
		// their plt entries (e.g. Mozilla and GhostScript), so check if there's
		// really a jmp_mem at this address
		// Still need to investigate further.
		if (GET_CHAR(target) == 0x68) {
			memcpy(X_instr,(char*)target,5);
			X_instr+=5;
			target+=5;
		}

		// don't instrument libpthread, it sometimes disables all signals ->
		// jiti gets disabled, can't instrument further and the whole program
		// hangs indefinitely
		new_target = GET_LONG(GET_LONG(target+2));
		if ((new_target >= libpthread_base) &&
		    (new_target <= libpthread_top)) {
			// adjust return address so it points to the clone
			new_ret_address = X_address+5+((X_address > end) ? ddelta:delta);
			// check whether we haven't instrumented this yet
			if ((target = LOOKUP_CLONE2INSTR(new_ret_address)) != 0)
				new_ret_address = target;
			else
				TO_BE_PATCHED(ret_address_address,X_address+5,TRUE);
			// overwrite the pushed return address
			PUT_LONG(ret_address_address,new_ret_address);
			// and jump to the uninstrumented routine
			JMPL(X_instr,new_target-(unsigned)X_instr-5);
			X_instr+=5;
			return 0;
		}

		if (_optimize_)  {
			return call_dyn_handler(target);
		}
	}

	if ((new_target = LOOKUP_CLONE2INSTR(target+curdelta))) {
		DBG(fprintf
				(jiti_tty,
				 "\n>>>Found target in clone2instr info! -> %x",
				 new_target));
		JMPL(X_instr, new_target - (unsigned) X_instr - 5);
		X_instr += 5;
		return 0;
	}

	if ((new_target = LOOKUP_RELOC(target))) {
		DBG(fprintf
				(jiti_tty, "\n>>>Found target in reloc info! -> %x",
				 new_target));
		JMPL(X_instr, new_target - (unsigned) X_instr - 5);
		X_instr += 5;
		return 0;
	}

	if (!FOLLOW_CALL_DIR || (follow++ > MAX_FOLLOW)) {
		X_instr_end_jmp = (unsigned)X_instr;

		// create a dummy jump that will be patched later. Can't just
		// patch the call after this, because that would require two
		// operations -> not thread-safe
		JMPL(X_instr,0);
		X_instr+=5;
		PUT_CALL(X_instr,jmpcall_dir_handler_ - (unsigned) X_instr - 5);
		X_instr+=5;
		PUT_LONG(X_instr,X_address);
		X_instr+=4;

		return 0;
	} else {
		DBG(fprintf(jiti_tty, "\n>>>Restart from %x", target));
		// NOTE: can't do it now yet, since otherwise someone else in RET_HANDLER could
		//   get this address and jump to it before we have generated the necessary
		//   code!!!!!
		SCHEDULE_UPDATE_CLONE2INSTR(target+curdelta,(unsigned)X_instr);
		_X_start = target;
		last_clone_link = target + curdelta;
		DBG(fprintf(jiti_tty,"Following, last clone link = 0x%x\n",last_clone_link));
		return target;
	}
}

//////////////////////////////////////////////////////

unsigned call_reg_handler(unsigned dummy1, unsigned dummy2, unsigned dummy3){

	if (lock)
		panic("lock before call_reg??");

	// traced in jmp_reg_handler
//	TRACE_CODE(_X_start, X_address + 1);

	/* return address */
	PUSHL32(X_instr, X_address + 2);
	X_instr += 5;
#ifdef MEM
	// traced in jmp_reg_handelr()
#endif
	return jmp_reg_handler(FALSE, DUMMY, DUMMY);

}

//////////////////////////////////////////////////////

unsigned call_mem_handler(unsigned len, unsigned dummy2, unsigned dummy3){

	if (lock)
		panic("lock before call_mem!!");

	// traced in jmp_mem_handler(
//	TRACE_CODE(_X_start, X_address);

	/* return address */
	PUSHL32(X_instr, X_address + len);
	X_instr += 5;

#ifdef MEM
	// traced in jmp_mem_handler()
#endif

	return jmp_mem_handler(len, FALSE, DUMMY);
}

//////////////////////////////////////////////////////

unsigned jcc_taken_handler(unsigned address){
	unsigned new_target, jcc_instr =
		*(unsigned *) (address), patch_address = address - 5, curdelta, jecxz;
	unsigned target;
	signed char disp8;

	jecxz = FALSE;
	// bereken target & condcode
	if (GET_CHAR(jcc_instr) != 0x0F) {	// disp8
		jecxz = (GET_CHAR(jcc_instr) == 0xe3);
		disp8 = GET_CHAR(jcc_instr + 1);
		target = jcc_instr + (signed) disp8 + 2;
	} else {			// disp32
		target = jcc_instr + GET_LONG(jcc_instr + 2) + 6;
	}


	DBG(fprintf(jiti_tty, "\n>>>In jcc_taken_handler(), coming from %x, "
				"ori jcc @ %x, target=%x, patch address=%X",
				address, jcc_instr, target, patch_address));

	curdelta = (target > end) ? ddelta : delta;

	new_target = LOOKUP_CLONE2INSTR(target+curdelta);
	DBG(fprintf(jiti_tty, "\n>>> JCC_TAKEN: LOOKUP_CLONE2INSTR(%x) -> %x", target+curdelta, new_target));

/* not thread safe!!! Also doesn't seem to produce many, if any, matches
	if (GEBRUIK_RELOC_INFO) {
		if (!new_target)
			new_target = LOOKUP_RELOC(target);
		DBG(fprintf(jiti_tty, "\n>>> JCC_TAKEN: LOOKUP_RELOC(%x) -> %x", target, new_target));
	}
*/

	// decodeer
	if (!new_target) {
		new_target = decode_block(target + curdelta);
	}
	// patch address of generated jcc  instr
	if (!jecxz) {
		patch_address = patch_address - 2 - 4;
		PUT_LONG(patch_address,new_target - patch_address - 4);
		DBG(patch_address -= 2;)
	}
/*
	else {
		// FIXME!!! Locking the decoder here is not enough. If another
		// thread is executing the call to this handler while we
		// change it (and this is possible, since MAKE_LINK requires
		// two operations), we can get unpredictable results.
		LOCK_DECODER();
		MAKE_LINK(patch_address, new_target);
		UNLOCK_DECODER();
	}
*/

	DBG(jiti_fprint_ins1(jiti_tty, 1, patch_address));

	DBG(fprintf
			(jiti_tty, "\n>>>DONE (jcc_taken_handler), restarting at 0x%x",
			 new_target));

	FLUSH_INSTRUCTIONS;

	// ok
	return new_target;
}

//////////////////////////////////////////////////////

unsigned jcc_handler(unsigned len, unsigned dummy1, unsigned dummy2){
	unsigned target, new_target, condcode, jecxz;
	signed char disp8;

	if (lock)
		panic("lock before jcc??");

	jecxz = FALSE;
	// bereken target & condcode
	if (len == 2) {
	
		disp8 = GET_CHAR(X_address + 1);
		target = X_address + (signed) disp8 + 2;

		// jecx?
		if (GET_CHAR(X_address) != 0xe3) {
			// no
			condcode = GET_CHAR(X_address) & 0x0F;
		} else {
			// yes
			jecxz = TRUE;
			condcode = 4; // (equal/zero)
		}

		new_target = (X_address > end)? X_address+ddelta:X_address+delta;
		
		// only replicate the jcc in the clone if there is no breakpoint at
		// the target address, because otherwise it's possible we end up
		// at this breakpoint every single time this jcc is taken (we won't
		// reinstrument it every time since clone2instr will cache the
		// corresponding X_instr code, but we will always take the breakpoint
		// penalty)
		if ((!_instrument_code_) &&
		    (GET_CHAR(new_target) != BP)) {
			memcpy((char *) new_target, (char *) X_address, 2);
		}
		else {
			CLEAN_CLONE_UNTIL(X_address + ((X_address > end) ? ddelta: delta));
		}

	} else {
		target = X_address + GET_LONG(X_address + 2) + 6;
		condcode = GET_CHAR(X_address + 1) & 0x0F;
		if (X_address > end) {
			SCHEDULE_MAKE_LINK(X_address + ddelta, X_instr_start_block);
		} else {
			SCHEDULE_MAKE_LINK(X_address + delta, X_instr_start_block);
		}
	}

	TRACE_CODE(_X_start, X_address + len - 1);
	_X_start = X_address + len;

	DBG(fprintf(jiti_tty, "\n>>>TARGET=%x/condcode=%d/", target, condcode));

	if (jecxz) {
		PUSHF;
		X_instr++;
		PUT_SHORT(X_instr,0xc985);
		X_instr+=2;	// testl %ecx,%ecx 
	}

	if (jecxz == FALSE) {
		new_target = LOOKUP_CLONE2INSTR(target+ ((target > end) ? ddelta:delta));
		if (!new_target)
			new_target = LOOKUP_RELOC(target);
		if (new_target) {
			target = new_target;
			PUT_CHAR(X_instr, 0x0F);
			X_instr++;
			PUT_CHAR(X_instr, 0x80 | condcode);
			X_instr++;
			PUT_LONG(X_instr, target - (unsigned) X_instr - 4);
			X_instr += 4;
			return X_address + len;
		}
		else {

			// only try patching for len == 2 since targets of longer jcc's most likely
			// won't be patchable -> don't waste time trying to
			if ((GEBRUIK_PATCH_INFO) &&
			    (len == 2))
				TO_BE_PATCHED(X_instr + 2, target,FALSE);

			if (target > end)
				target += ddelta;
			else
				target += delta;
		}
		DBG(fprintf(jiti_tty, "\n>>>newTARGET=%x ", target));
	} else {
		if (target > end)
			target += ddelta;
		else
			target += delta;
	}

#if 0
	PUT_CHAR(X_instr, 0x0F);
	X_instr++;
	PUT_CHAR(X_instr, 0x80 | condcode);
	X_instr++;
	PUT_LONG(X_instr, target - (unsigned) X_instr - 4);
	X_instr += 4;
#else
	target = (unsigned) jcc_taken_handler_;
	DBG(fprintf(jiti_tty, "\n>>>LINKING TO jcc_taken_handler() at %x!",
			 jcc_taken_handler_));

	PUT_CHAR(X_instr, 0x0F);
	X_instr++;			// Jcc +6+ord(jecxz)
	PUT_CHAR(X_instr, 0x80 | condcode);
	X_instr++;
	PUT_LONG(X_instr, 2+jecxz);
	X_instr += 4;

	if (jecxz == TRUE) {
		POPF;
		X_instr++;
	}
	PUT_CHAR(X_instr, 0xEB);
	X_instr++;			// JMP +9+ord(jecxz)
	PUT_CHAR(X_instr, 9+jecxz);
	X_instr++;

	if (jecxz == TRUE) {
		POPF;
		X_instr++;
	}
	PUT_CALL(X_instr, jcc_taken_handler_ - (unsigned) X_instr - 5);
	X_instr += 5;		// CALL jcc_taken_handler

	PUT_LONG(X_instr, X_address);
	X_instr += 4;		// adres van ori. jcc
#endif

	return X_address + len;

}

//////////////////////////////////////////////////////

unsigned pusha_popa_handler(unsigned type, unsigned dummy1, unsigned dummy2) {

	STACK_ACCESS(type,32);
	ORIGINAL_INSTRUCTION(1);
	return X_address;
}

unsigned btx_handler(unsigned type, unsigned dummy1, unsigned dummy2) {

	unsigned memlen;
	char srcreg, calcreg;

	if (byte2 >> 6 == 3) {
		// INSTR_LEN(2) because the 0x0f (and maybe other) prefix(es) will be
		// added by no_mem/ORIGINAL_INSTRUCTION
		return no_mem(INSTR_LEN(2),IMM_LEN(0),DUMMY);
	}

	memlen = get_mem_len(byte2);

	/* "btx reg, mem" can address a bit position within the range -2^31 to 
	 * 2^31 - 1 (bits) of mem -> we have to calculate the offset every
	 * time if we are tracing the memory accesses
	 */
	if (_instrument_mem_) {

		if (immsize_override)
			panic("TODO: size override support for btx!!");

		CLEAN_CLONE_UNTIL(X_address + ((X_address > end) ? ddelta: delta));

		srcreg = (GET_CHAR(X_address+prefix_c+1) & ~0xc7) >> 3;

		// we need two different registers
		if (srcreg != EAX)
			calcreg = EAX;
		else
			calcreg = EDX;

		PUSHF;
		X_instr++;       // save flags

		PUSH_REG(srcreg);
		X_instr++;
		PUSH_REG(calcreg);
		X_instr++;       // save registers used for address calculation

		// load address of target in calcreg using a lea instruction
		// note: if the memory expression uses esp, the calculated value should be
		// adjusted!!
		PUT_CHAR(X_instr, 0x8D);	// lea
		memcpy((void *) X_instr+1, (void *) X_address+prefix_c+1, memlen+1);	// copy address expression
		PUT_CHAR(X_instr + 1, (GET_CHAR(X_instr + 1) & 0xC7) | (calcreg << 3));	// ,%calcreg
		X_instr+=memlen+2;

		// sar by 3 to get the byte in which the bit is modified
		// note: this is not a normal division by 8: -1 sar 3 = -1, but this is what
		//       we want: bit -1 relative to address addr lies in the byte at addr-1

		PUT_CHAR(X_instr, 0xc1);
		X_instr++;                                       // sarl
		PUT_CHAR(X_instr, (0x1f << 3) | srcreg);
		X_instr++;                                       // , %srcreg
		PUT_CHAR(X_instr, 3);
		X_instr++;                                       // $3

		// mask out the two lower bits, since the processor will access the int
		// at address+(4*(bitoffset div 32)) (page 3-45 of the "Intel Architecture
		// Software Developer's Manual, Volume 2: Instruction Set Reference"

		PUT_CHAR(X_instr,0x83);
		X_instr++;
		PUT_CHAR(X_instr,0xe0 | srcreg);
		X_instr++;
		PUT_CHAR(X_instr,~3);
		X_instr++;                                       // andl $~3,%srcreg

		// add it to the address
		PUT_CHAR(X_instr,0x1);                           // add
		X_instr++;
		PUT_CHAR(X_instr,0xc0 | (srcreg << 3) | calcreg);
		X_instr++;                                       // %srcreg,%calcreg

		PUSH_REG(calcreg);
		X_instr++;  // push the resulting address on the stack

		PUSHL8(X_instr,4);
		X_instr += 2;  // this instruction accesses 4 bytes, see above

		PUSHL8(X_instr,type)
		X_instr += 2;  // the kind of access

		PUSHL32(X_instr,X_address);
		X_instr += 5; // the address of the instruction

		PUT_CALL(X_instr, trace_ - (unsigned) X_instr - 5);
		X_instr += 5; // trace

		// restore saved registers and flags
		POP_REG(calcreg);
		X_instr++;

		POP_REG(srcreg);
		X_instr++;

		POPF;
		X_instr++;

	} else {
	  X_kloon = 1;
	}

	ORIGINAL_INSTRUCTION(2+memlen);
	return X_address;
}

//////////////////////////////////////////////////////

void rep_handler(unsigned si_type, unsigned di_type, unsigned X_len){

	/* voor de instructie */
	if (di_type != REP_NONE) {
		PUSHL32(X_instr, X_address);
		X_instr += 5;
		PUSH_REG(EDI);
		X_instr++;
	}

	if (si_type != REP_NONE) {
		PUSHL32(X_instr, X_address);
		X_instr += 5;
		PUSH_REG(ESI);
		X_instr++;
	}

	/* de instructie zelf */
	memcpy((void *) X_instr, (void *) X_address, X_len);
	X_instr += X_len;

	/* na de instructie */
	if (si_type == REP_LOAD) {	/* REP_STORE kan niet! */
		PUSH_REG(ESI);
		X_instr++;
		PUSHL8(X_instr, TYPE(LOAD));
		X_instr += 2;
		PUT_CALL(X_instr, range - (unsigned) X_instr - 5);
		X_instr += 5;
	}

	if (di_type != REP_NONE) {
		PUSH_REG(EDI);
		X_instr++;
		PUSHL8(X_instr, di_type == REP_LOAD ? TYPE(LOAD) : TYPE(STORE));
		X_instr += 2;
		PUT_CALL(X_instr, range - (unsigned) X_instr - 5);
		X_instr += 5;
	}

}

///////////////////////////////////////////////////////

unsigned rep(unsigned dummy1, unsigned dummy2, unsigned dummy3){
	unsigned si_type = 0, di_type = 0, X_len = 0;

	if (lock)
		panic("lock before rep??");

	if ((byte1 == 0xF2) || (byte1 == 0xF3)) {	// REPNE of REP/REPE
		X_len++;
		byte1 = byte2;
		byte2 = byte3;
		byte3 = byte4;
	}

	if (byte1 == 0x66) {	// OPERAND SIZE PREFIX
		X_len++;
		byte1 = byte2;
		byte2 = byte3;
		byte3 = 0;
	}

	switch (byte1) {
		case 0xA6:			/* REPNE CMPSb */
			si_type = REP_LOAD, di_type = REP_LOAD;
			break;
		case 0xA7:			/* REPNE CMPSw */
			si_type = REP_LOAD, di_type = REP_LOAD;
			break;
		case 0xAE:			/* REPNE SCASb */
			si_type = REP_NONE, di_type = REP_LOAD;
			break;
		case 0xAF:			/* REPNE SCASw */
			si_type = REP_NONE, di_type = REP_LOAD;
			break;
		case 0x6C:			/* REP INSb    */
			panic("REP PORT!");
		case 0x6D:			/* REP INSw    */
			panic("REP PORT!");
		case 0xAC:			/* REP LODSb   */
			si_type = REP_LOAD, di_type = REP_NONE;
			break;
		case 0xAD:			/* REP LODSw   */
			si_type = REP_LOAD, di_type = REP_NONE;
			break;
		case 0xA4:			/* REP MOVSb   */
			si_type = REP_LOAD, di_type = REP_STORE;
			break;
		case 0xA5:			/* REP MOVSw   */
			si_type = REP_LOAD, di_type = REP_STORE;
			break;
		case 0x6E:			/* REP OUTSb   */
			panic("REP PORT!");
		case 0x6F:			/* REP OUTSw   */
			panic("REP PORT!");
		case 0xAA:			/* REP STOSb   */
			si_type = REP_NONE, di_type = REP_STORE;
			break;
		case 0xAB:			/* REP STOSw   */
			si_type = REP_NONE, di_type = REP_STORE;
			break;
		case 0x90:
			// 'repz nop' appears in some versions of libpthread
			return no_mem(INSTR_LEN(X_len+1),IMM_LEN(0), DUMMY);
			break;
		default:
			panic("Unknown rep!");
	}
	// count opcode
	X_len++;
	if (0 /*_instrument_mem_*/ ) {
		rep_handler(si_type, di_type,X_len);
		return X_address + X_len;
	} else {
		X_kloon = 1;
		ORIGINAL_INSTRUCTION(X_len);
		return X_address;
	}

}

/////////////////////////////////////////////////////////////

unsigned opcodeFF(unsigned dummy1, unsigned dummy2, unsigned dummy3){
	unsigned char opcode2 = (byte2 & 0x3F) >> 3;
	unsigned char mod2 = byte2 >> 6;

	switch (opcode2) {
		case 0:		/* DEC of INC MEM */
		case 1:		/* DEC of INC MEM */
			return check_mod(IMM_LEN(0), MEM_LEN(4), TYPE(MODIFY));
		case 2:
		case 3:
			if (mod2 == 3) {	/* CALL *REG */
				call_reg_handler(DUMMY, DUMMY, DUMMY);
				return 0;
			} else {		/* CALL *MEM */
				call_mem_handler(2 + get_mem_len(byte2), DUMMY, DUMMY);
				return 0;
			}
		case 4:
		case 5:
			if (mod2 == 3) {	/* JMP *REG */
				jmp_reg_handler(TRUE, DUMMY, DUMMY);
				return 0;
			} else {		/* JMP *MEM */
				jmp_mem_handler(2 + get_mem_len(byte2), TRUE, DUMMY);
				return 0;
			}
		case 6:			/* PUSH *MEM */
			return push_pop_mem_handler(TYPE(STORE), TYPE(LOAD), DUMMY);
		case 7:
			return dontexist(DUMMY, DUMMY, DUMMY);	/* illegal */
	}
	panic("assert");
	return -1;			// not reached
}

//////////////////////////////////////////////////////////////////////////

typedef struct decode {
	unsigned (*function) (unsigned, unsigned, unsigned);
	unsigned arg1;
	unsigned arg2;
	unsigned arg3;
} decode;

///////////////////////////////////////////////////////////////////////////

/* TABELLEN */

decode opcode_table[256] = {
	/* 0x00 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(MODIFY)}, /* ADD */
	/* 0x01 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(MODIFY)}, /* ADD */
	/* 0x02 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* ADD */
	/* 0x03 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* ADD */

	/* 0x04 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* ADD IMM8,A? */
	/* 0x05 */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* ADD IMM32,A? */

	/* 0x06 */ {dontexist},
	/* 0x07 */ {dontexist},

	/* 0x08 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* OR */
	/* 0x09 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* OR */
	/* 0x0a */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* OR */
	/* 0x0b */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* OR */

	/* 0x0c */ {no_mem, INSTR_LEN(2),IMM_LEN(1)}, /* OR IMM,AL */
	/* 0x0d */ {no_mem, INSTR_LEN(5),IMM_LEN(4)}, /* OR IMM,EAX */

	/* 0x0e */ {dontexist},

	/* 0x0f */ {extended0F}, /* Jcc, MMX, SSE */

	/* 0x10 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* ADC */
	/* 0x11 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* ADC */
	/* 0x12 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* ADC */
	/* 0x13 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* ADC */

	/* 0x14 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* ADC IMM8, AL */
	/* 0x15 */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* ADC IMM32, AX */

	/* 0x16 */ {dontexist},
	/* 0x17 */ {dontexist},

	/* 0x18 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* SBB */
	/* 0x19 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* SBB */
	/* 0x1a */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* SBB */
	/* 0x1b */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* SBB */

	/* 0x1c */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* SBB IMM8, AL */
	/* 0x1d */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* SBB IMM32, AL */

	/* 0x1e */ {dontexist},
	/* 0x1f */ {dontexist},

	/* 0x20 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* AND */
	/* 0x21 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* AND */
	/* 0x22 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* AND */
	/* 0x23 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* AND */

	/* 0x24 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* AND IMM8, AL */
	/* 0x25 */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* AND IMM32, AL */

	/* 0x26 */ {unsupported}, /* ES override, not used in Linux */

	/* 0x27 */ {dontexist},

	/* 0x28 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* SUB */
	/* 0x29 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* SUB */
	/* 0x2a */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* SUB */
	/* 0x2b */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* SUB */

	/* 0x2c */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* SUB IMM8,AL */
	/* 0x2d */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* SUN IMM32,AL */

	/* 0x2e */ {unsupported}, /* CS override, not used in Linux */

	/* 0x2f */ {dontexist},

	/* 0x30 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* XOR */
	/* 0x31 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* XOR */
	/* 0x32 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* XOR */
	/* 0x33 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* XOR */

	/* 0x34 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* XOR IMM8,AL */
	/* 0x35 */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* XOR IMM32,EAX */

	/* 0x36 */ {unsupported}, /* SS override, not used in Linux */

	/* 0x37 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* AAA */

	/* 0x38 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* CMP mem8,r  of CMP r1,r2 */
	/* 0x39 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* CMP mem32,r of CMP r1,r2 */
	/* 0x3a */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* CMP r,mem8  of CMP r2,r1 */
	/* 0x3b */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* CMP r,mem32 of CMP r2,r1 */

	/* 0x3c */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* CMP IMM8, AL */
	/* 0x3d */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* CMP IMM32, AL */

	/* 0x3e */ {unsupported}, /* DS override, not used in Linux */

	/* 0x3f */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* AAS */

	/* 0x40 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */
	/* 0x41 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */
	/* 0x42 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */
	/* 0x43 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */
	/* 0x44 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */
	/* 0x45 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */
	/* 0x46 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */
	/* 0x47 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* INC REG */

	/* 0x48 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */
	/* 0x49 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */
	/* 0x4a */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */
	/* 0x4b */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */
	/* 0x4c */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */
	/* 0x4d */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */
	/* 0x4e */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */
	/* 0x4f */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* DEC REG */

	/* 0x50 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL EAX, 1 byte */
	/* 0x51 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL ECX */
	/* 0x52 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL EDX */
	/* 0x53 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL EBX, SSE?? [TODO] */
	/* 0x54 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL ESP */
	/* 0x55 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL EBP */
	/* 0x56 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL ESI */
	/* 0x57 */ {push_pop_handler, INSTR_LEN(1), TYPE(STORE)}, /* PUSHL EDI */
	/* 0x58 */ {push_pop_handler, INSTR_LEN(1), TYPE(LOAD)}, /* POPL  EAX */
	/* 0x59 */ {push_pop_handler, INSTR_LEN(1), TYPE(LOAD)}, /* POPL  ECX */
	/* 0x5a */ {push_pop_handler, INSTR_LEN(1), TYPE(LOAD)}, /* POPL  EDX */
	/* 0x5b */ {push_pop_handler, INSTR_LEN(1), TYPE(LOAD)}, /* POPL  EBX */
	/* 0x5c */ {dontexist}, /* POPL  ESP */
	/* 0x5d */ {push_pop_handler, INSTR_LEN(1), TYPE(LOAD)}, /* POPL  EBP */
	/* 0x5e */ {push_pop_handler, INSTR_LEN(1), TYPE(LOAD)}, /* POPL  ESI */
	/* 0x5f */ {push_pop_handler, INSTR_LEN(1), TYPE(LOAD)}, /* POPL  EDI */

	/* 0x60 */ {pusha_popa_handler,TYPE(STORE)}, /* pusha */
	/* 0x61 */ {pusha_popa_handler,TYPE(LOAD)},  /* popa */
	/* 0x62 */ {dontexist},

	/* 0x63 */ {privileged}, /* ARPL */

	/* 0x64 */ {unsupported}, /* FS override, not used in Linux */
	/* 0x65 */ {PSegOverride, GS}, /* GS override, used in libpthread?? */

	/* 0x66 */ {Poperand}, /* prefix: operand size override */
	/* 0x67 */ {Paddress}, /* prefix: address size override */

	/* 0x68 */ {push_pop_handler, INSTR_LEN(5), TYPE(STORE)}, /* PUSHL IMM32, 5 bytes */

	/* 0x69 */ {check_mod, IMM_LEN(4), MEM_LEN(4), TYPE(LOAD)}, /* imull reg1,simm8,reg2 */

	/* 0x6a */ {push_pop_handler, INSTR_LEN(2), TYPE(STORE)}, /* PUSHL IMM8, 2 bytes */

	/* 0x6b */ {check_mod, IMM_LEN(1), MEM_LEN(1), TYPE(LOAD)}, /* imull reg1,imm32,reg2 */
	
	/* 0x6c */ {dontexist},
	/* 0x6d */ {dontexist},

	/* 0x6e */ {privileged}, /* OUTS */
	/* 0x6f */ {privileged}, /* OUTS */

	/* 0x70 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x71 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x72 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x73 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x74 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x75 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x76 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x77 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x78 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x79 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x7a */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x7b */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x7c */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x7d */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x7e */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */
	/* 0x7f */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0) /* Jcc no_mem */ },	/* Jcc 8b, 2 bytes */

	/* 0x80 */ {check_mod_op2a, IMM_LEN(1), MEM_LEN(1)}, /* MOV,CMP,ADD,XOR imm,r of imm,mem, met  */
	/* 0x81 */ {check_mod_op2a, IMM_LEN(4), MEM_LEN(4)}, /* MOV,CMP,ADD,XOR imm,r of imm,mem, met w */
	/* 0x82 */ {check_mod_op2a, IMM_LEN(1), MEM_LEN(1)}, /* MOV,CMP,ADD,XOR imm,r of imm,mem, met s */
	/* 0x83 */ {check_mod_op2a, IMM_LEN(1), MEM_LEN(4)}, /* MOV,CMP,ADD,XOR imm,r of imm,mem, met sw */

	/* 0x84 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* TEST */
	/* 0x85 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* TEST */

	/* 0x86 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(MODIFY)}, /* XCHG */
	/* 0x87 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(MODIFY)}, /* XCHG */

	/* 0x88 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* MOV r1,r2 of MOV r,mem  8b */
	/* 0x89 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* MOV r1,r2 of MOV r,mem 32b */
	/* 0x8a */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)}, /* MOV r2,r1 of MOV mem,r  8b */
	/* 0x8b */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* MOV r2,r1 of MOV mem,r 32b */

	/* 0x8c */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* MOV ?S,r/m16 */

	/* 0x8d */ {no_mem, INSTR_LEN(CALC), IMM_LEN(0)}, /* LEAL */

	/* 0x8e */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)}, /* MOV r/m16,?S */

	/* 0x8f */ {push_pop_mem_handler, TYPE(LOAD), TYPE(STORE)}, /* pop mem, used in setjmp procedure of FPC */

	/* 0x90 */ {nop, INSTR_LEN(1)}, /* nop */

	/* 0x91 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* xchg al,reg */
	/* 0x92 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* xchg al,reg */
	/* 0x93 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* xchg al,reg */
	/* 0x94 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* xchg al,reg */
	/* 0x95 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* xchg al,reg */
	/* 0x96 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* xchg al,reg */
	/* 0x97 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* xchg al,reg */

	/* 0x98 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* CBW */
	/* 0x99 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* CDQ */

	/* 0x9a */ {dontexist},

	/* 0x9b */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* FWAIT */

	/* 0x9c */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* pushf: TODO */
	/* 0x9d */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* popf:  TODO */

	/* 0x9e */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* SAHF */

	/* 0x9f */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, // TODO: LAHF

	/* 0xa0 */ {movl_eax_mem_handler, INSTR_LEN(5), MEM_LEN(1), TYPE(LOAD)}, /* MOV MEM, AL */
	/* 0xa1 */ {movl_eax_mem_handler, INSTR_LEN(5), MEM_LEN(4), TYPE(LOAD)}, /* MOV MEM, EAX */
	/* 0xa2 */ {movl_eax_mem_handler, INSTR_LEN(5), MEM_LEN(1), TYPE(STORE)}, /* MOV AL,  MEM */
	/* 0xa3 */ {movl_eax_mem_handler, INSTR_LEN(5), MEM_LEN(4), TYPE(STORE)}, /* MOV EAX, MEM */

	/* 0xa4 */ {rep}, /* movsb */
	/* 0xa5 */ {rep}, /* movsw */
	/* 0xa6 */ {rep /*restart, cmpsb_handler */ },	/* cmpsb */
	/* 0xa7 */ {rep}, /* cmpsw */

	/* 0xa8 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* TEST IMM8,AL */
	/* 0xa9 */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* TEST IMM32, EAX */

	/* 0xaa */ {rep}, /* stosb */
	/* 0xab */ {rep}, /* stosw */
	/* 0xac */ {rep}, /* lodsb */
	/* 0xad */ {rep}, /* lodsw */
	/* 0xae */ {rep}, /* scasb */
	/* 0xaf */ {todo}, /* scasw */

	/* 0xb0 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb1 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb2 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb3 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb4 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb5 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb6 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb8 */ {no_mem, INSTR_LEN(2), IMM_LEN(1)}, /* MOV IMM8,REG */
	/* 0xb9 */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */
	/* 0xb7 */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */
	/* 0xba */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */
	/* 0xbb */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */
	/* 0xbc */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */
	/* 0xbd */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */
	/* 0xbe */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */
	/* 0xbf */ {no_mem, INSTR_LEN(5), IMM_LEN(4)}, /* MOV IMM32,REG */

	/* 0xc0 */ {check_mod, IMM_LEN(1), MEM_LEN(1)}, /* RCL/RCR/ROL/ROR/SAR IMM */
	/* 0xc1 */ {check_mod, IMM_LEN(1), MEM_LEN(4)}, /* TODO */
	
	/* 0xc2 */ {ret_handler, INSTR_LEN(3)}, /* ret imm, 3 bytes */

	/* 0xc3 */ {ret_handler, INSTR_LEN(1)},	/* ret, 1 byte */

	/* 0xc4 */ {dontexist},
	/* 0xc5 */ {dontexist},

	/* 0xc6 */ {check_mod, IMM_LEN(1), MEM_LEN(1), TYPE(STORE)}, /* MOV IMM,REG */
	/* 0xc7 */ {check_mod, IMM_LEN(4), MEM_LEN(4), TYPE(STORE)}, /* MOV IMM,REG */

	/* 0xc8 */ {dontexist}, /* enter */
	/* 0xc9 */ {leave /*restart, leave_handler */ },	/* leave */

	/* 0xca */ {todo}, /* ret imm, other seg */
	/* 0xcb */ {todo}, /* ret, other seg */

	/* 0xcc */ {dontexist},
	/* 0xcd */ {interrupt_handler}, /* interrupt, appears when calling the kernel */
	/* 0xce */ {dontexist},
	/* 0xcf */ {dontexist},

	/* 0xd0 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* RCR/RCL/ROL/ROR/SAR: 1 */
	/* 0xd1 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)},
	/* 0xd2 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* : CL */
	/* 0xd3 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(STORE)}, /* : CL */

	/* 0xd4 */ {no_mem, INSTR_LEN(2), IMM_LEN(0)}, /* AAM */
	/* 0xd5 */ {no_mem, INSTR_LEN(2), IMM_LEN(0)}, /* AAD */

	/* 0xd6 */ {dontexist},
	/* 0xd7 */ {todo}, /* XLAT */

	/* FLOATING POINT START **************** */
	/* 0xd8 */ {_fp}, /* altijd 32bit toegang */
	/* 0xd9 */ {_fp}, /* altijd 32bit toegang */
	/* 0xda */ {_fp}, /* altijd 32bit toegang */
	/* 0xdb */ {_fp}, /* altijd 32bit of 80bit toegang */
	/* 0xdc */ {_fp}, /* altijd 64bit toegang */
	/* 0xdd */ {_fp}, /* altijd 64bit toegang */
	/* 0xde */ {_fp}, /* altijd 16bit toegang */
	/* 0xdf */ {_fp}, /* altijd 16bit of 64bit toegang */
	/* FLOATING POINT STOP **************** */

	/* 0xe0 */ {loop}, /* LOOPNZ */
	/* 0xe1 */ {loop}, /* LOOPZ */
	/* 0xe2 */ {loop}, /* LOOP */

	/* 0xe3 */ {jcc_handler, INSTR_LEN(2), IMM_LEN(0)},
	/* 0xe4 */ {no_mem,INSTR_LEN(2)}, /* IN PORT imm, reg8 */
	/* 0xe5 */ {no_mem,INSTR_LEN(2)}, /* IN PORT imm, reg32 */

	/* 0xe6 */ {no_mem,INSTR_LEN(2)}, /* OUT PORT reg8, imm */
	/* 0xe7 */ {no_mem,INSTR_LEN(2)}, /* OUT PORT reg32, imm */

	/* 0xe8 */ {call_dir_handler}, /* CALL DIRECT, 5 bytes */

	/* 0xe9 */ {jmp_dir_handler, INSTR_LEN(5) /*jmp32b no_mem, 5 */ },	/* JMP 32b, 5 bytes */

	/* 0xea */ {dontexist},

	/* 0xeb */ {jmp_dir_handler, INSTR_LEN(2) /*jmp8b no_mem, 2 */ },	/* JMP 8b,2 bytes */

	/* 0xec */ {no_mem,INSTR_LEN(1)},	/* IN PORT reg, reg8 */
	/* 0xed */ {no_mem,INSTR_LEN(1)}, /* IN PORT reg, reg32 */

	/* 0xee */ {no_mem,INSTR_LEN(1)}, /* OUT PORT reg, reg8 */
	/* 0xef */ {no_mem,INSTR_LEN(1)}, /* OUT PORT reg, reg32 */

	/* 0xf0 */ {Plock}, /* LOCK prefix */

	/* 0xf1 */ {dontexist},

	/* 0xf2 */ {rep, 0}, /* REPNE */
	/* 0xf3 */ {rep, 1}, /* REPE/REP, SSE[TODO] */

	/* 0xf4 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* HLT */

	/* 0xf5 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* CMC */

	/* 0xf6 */ {check_mod_op2b, IMM_LEN(1), MEM_LEN(1), TYPE(LOAD)}, /* DIV,MUL,NEG,NOT,TEST */
	/* 0xf7 */ {check_mod_op2b, IMM_LEN(4), MEM_LEN(4), TYPE(LOAD)}, /* DIV,MUL,NEG,NOT,TEST */

	/* 0xf8 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* CLC */
	/* 0xf9 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* STC */
	/* 0xfa */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* CLI */
	/* 0xfb */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* STI */
	/* 0xfc */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* CLD */
	/* 0xfd */ {no_mem, INSTR_LEN(1), IMM_LEN(0)}, /* STD */

	/* 0xfe */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)}, /* inc mem8, dec mem8 */
	/* 0xff */ {opcodeFF} /* inc mem32, dec mem32, PUSH MEM, call *mem, jmp *mem */
};

void X_fetch(unsigned addr){
	fetch_address = addr;
	prefix_c = 0;
	immsize_override = 0;
	byte1 = GET_CHAR(fetch_address);
	byte2 = GET_CHAR(fetch_address + 1);
	byte3 = GET_CHAR(fetch_address + 2);
	byte4 = GET_CHAR(fetch_address + 3);
}

void skip(){
	prefix_c++;
	byte1 = byte2;
	byte2 = byte3;
	byte3 = byte4;
	byte4 = 0;
}

///////////////////////////////////////////////

unsigned Poperand(){

	switch (byte2) {
		case 0xA0:
			return todo(DUMMY, DUMMY, DUMMY);
		case 0xA1:
			prefix_c++;
			return movl_eax_mem_handler(INSTR_LEN(5), MEM_LEN(2), TYPE(LOAD));
		case 0xA2:
			return todo(DUMMY, DUMMY, DUMMY);
		case 0xA3:
			prefix_c++;
			return movl_eax_mem_handler(INSTR_LEN(5), MEM_LEN(2), TYPE(STORE));
	}

	skip();

	immsize_override = 2;
	return opcode_table[byte1].function(opcode_table[byte1].arg1,
			opcode_table[byte1].arg2,
			opcode_table[byte1].arg3);
}

///////////////////////////////////////////////

unsigned Plock(){

	lock = 1;
	skip();

	return opcode_table[byte1].function(opcode_table[byte1].arg1,
			opcode_table[byte1].arg2,
			opcode_table[byte1].arg3);
}

///////////////////////////////////////////////

unsigned PSegOverride(unsigned reg, unsigned dummy1, unsigned dummy2) {

	seg_override = reg;
	skip();

	return opcode_table[byte1].function(opcode_table[byte1].arg1,
			opcode_table[byte1].arg2,
			opcode_table[byte1].arg3);

}

///////////////////////////////////////////////


decode opcode0F_table[256] = {
	/* 0x00 */ {privileged},				/* LIDT, LLDT, LTR, STR, VERR, VERW, SLDT */
	/* 0x01 */ {privileged},				/* INVLPG, LGDT, LMSW, SMSW */

	/* 0x02 */ {privileged},				/* LAR */
	/* 0x03 */ {privileged},				/* LSL */

	/* 0x04 */ {dontexist},
	/* 0x05 */ {dontexist},

	/* 0x06 */ {privileged},				/* CLTS */

	/* 0x07 */ {dontexist},

	/* 0x08 */ {privileged},				/* INVD */
	/* 0x09 */ {privileged},				/* WBINV */

	/* 0x0a */ {dontexist},

	/* 0x0b */ {dontexist},				/* UB2 */

	/* 0x0c */ {dontexist},

	/* 0x0d */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* 3DNow: PREFETCH */
	/* 0x0e */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* 3DNow: FEMMS */
	/* 0x0f */ {_3dnow},				/* 3DNow */

	/* 0x10 */ {_sse},				/* MOVUPS */
	/* 0x11 */ {_sse},				/* MOVUPS */
	/* 0x12 */ {_sse},				/* MOVHLPS, MOVLPS */
	/* 0x13 */ {_sse},				/* MOVLPS */
	/* 0x14 */ {_sse},				/* UNPCKLPS */
	/* 0x15 */ {_sse},				/* UNPCKHPS */
	/* 0x16 */ {_sse},				/* MOVHPS, MOVLHPS */
	/* 0x17 */ {_sse},				/* MOVHPS */
	/* 0x18 */ {_sse},				/* PREFETCH. */

	/* 0x19 */ {dontexist},
	/* 0x1a */ {dontexist},
	/* 0x1b */ {dontexist},
	/* 0x1c */ {dontexist},
	/* 0x1d */ {dontexist},
	/* 0x1e */ {dontexist},
	/* 0x1f */ {dontexist},

	/* 0x20 */ {no_mem, INSTR_LEN(2), IMM_LEN(0)},				/* MOV CR */
	/* 0x21 */ {no_mem, INSTR_LEN(2), IMM_LEN(0)},				/* MOV CR */
	/* 0x22 */ {no_mem, INSTR_LEN(2), IMM_LEN(0)},				/* MOV CR */
	/* 0x23 */ {no_mem, INSTR_LEN(2), IMM_LEN(0)},				/* MOV CR */

	/* 0x24 */ {dontexist},
	/* 0x25 */ {dontexist},
	/* 0x26 */ {dontexist},
	/* 0x27 */ {dontexist},

	/* 0x28 */ {_sse},				/* MOVAPS */
	/* 0x29 */ {_sse},				/* MOVAPS */
	/* 0x2a */ {_sse},				/* CVTPI2PS */
	/* 0x2b */ {_sse},				/* MOVNTPS */
	/* 0x2c */ {_sse},				/* CVTTPS2PI */
	/* 0x2d */ {_sse},				/* CVTPS2PI */
	/* 0x2e */ {_sse},				/* UCOMISS */
	/* 0x2f */ {_sse},				/* COMISS */

	/* 0x30 */ {privileged},				/* WRMSR */

	/* 0x31 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* RDTSC */
	/* 0x32 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* RDMSR */
	/* 0x33 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* RDPMC */

	/* 0x34 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* SYSENTER */
	/* 0x35 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* SYSEXIT */

	/* 0x36 */ {dontexist},
	/* 0x37 */ {dontexist},
	/* 0x38 */ {dontexist},
	/* 0x39 */ {dontexist},
	/* 0x3a */ {dontexist},
	/* 0x3b */ {dontexist},
	/* 0x3c */ {dontexist},
	/* 0x3d */ {dontexist},
	/* 0x3e */ {dontexist},
	/* 0x3f */ {dontexist},

	/* 0x40 */ {cmovcc},				/* CMOVcc */
	/* 0x41 */ {cmovcc},				/* CMOVcc */
	/* 0x42 */ {cmovcc},				/* CMOVcc */
	/* 0x43 */ {cmovcc},				/* CMOVcc */
	/* 0x44 */ {cmovcc},				/* CMOVcc */
	/* 0x45 */ {cmovcc},				/* CMOVcc */
	/* 0x46 */ {cmovcc},				/* CMOVcc */
	/* 0x47 */ {cmovcc},				/* CMOVcc */
	/* 0x48 */ {cmovcc},				/* CMOVcc */
	/* 0x49 */ {cmovcc},				/* CMOVcc */
	/* 0x4a */ {cmovcc},				/* CMOVcc */
	/* 0x4b */ {cmovcc},				/* CMOVcc */
	/* 0x4c */ {cmovcc},				/* CMOVcc */
	/* 0x4d */ {cmovcc},				/* CMOVcc */
	/* 0x4e */ {cmovcc},				/* CMOVcc */
	/* 0x4f */ {cmovcc},				/* CMOVcc */

	/* 0x50 */ {_sse},				/* MOVMSKPS */
	/* 0x51 */ {_sse},				/* SQRTPS */
	/* 0x52 */ {_sse},				/* RSQRTPS */
	/* 0x53 */ {_sse},				/* RCPPS */
	/* 0x54 */ {_sse},				/* ANDPS */
	/* 0x55 */ {_sse},				/* ANDNPS */
	/* 0x56 */ {_sse},				/* ORPS */
	/* 0x57 */ {_sse},				/* XORPS */
	/* 0x58 */ {_sse},				/* ADDPS */
	/* 0x59 */ {_sse},				/* MULPS */

	/* 0x5a */ {dontexist},
	/* 0x5b */ {dontexist},

	/* 0x5c */ {_sse},				/* SUBPS */
	/* 0x5d */ {_sse},				/* MINPS */

	/* 0x5e */ {dontexist},

	/* 0x5f */ {_sse},				/* MAXPS */

	/* 0x60 */ {_mmx},
	/* 0x61 */ {_mmx},				/* SSE: SQRTPS!!! */
	/* 0x62 */ {_mmx},
	/* 0x63 */ {_mmx},
	/* 0x64 */ {_mmx},
	/* 0x65 */ {_mmx},
	/* 0x66 */ {_mmx},
	/* 0x67 */ {_mmx},
	/* 0x68 */ {_mmx},
	/* 0x69 */ {_mmx},
	/* 0x6a */ {_mmx},
	/* 0x6b */ {_mmx},

	/* 0x6c */ {dontexist},
	/* 0x6d */ {dontexist},

	/* 0x6e */ {_mmx},
	/* 0x6f */ {_mmx},

	/* 0x70 */ {no_mem, INSTR_LEN(3), IMM_LEN(1)},				/* MMX met imm8 *//* SSE: PSHUFW!! */
	/* 0x71 */ {no_mem, INSTR_LEN(3), IMM_LEN(1)},				/* MMX met imm8 */
	/* 0x72 */ {no_mem, INSTR_LEN(3), IMM_LEN(1)},				/* MMX met imm8 */
	/* 0x73 */ {no_mem, INSTR_LEN(3), IMM_LEN(1)},				/* MMX met imm8 */

	/* 0x74 */ {_mmx},
	/* 0x75 */ {_mmx},
	/* 0x76 */ {_mmx},

	/* 0x77 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* MMX: EMMS */

	/* 0x78 */ {dontexist},
	/* 0x79 */ {dontexist},
	/* 0x7a */ {dontexist},
	/* 0x7b */ {dontexist},
	/* 0x7c */ {dontexist},
	/* 0x7d */ {dontexist},

	/* 0x7e */ {_mmx},
	/* 0x7f */ {_mmx},

	/* 0x80 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x81 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x82 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x83 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x84 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x85 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x86 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x87 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x88 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x89 */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x8a */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x8b */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x8c */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x8d */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x8e */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */
	/* 0x8f */ {jcc_handler, INSTR_LEN(6)},				/* Jcc IMM32 */

	/* 0x90 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x91 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x92 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x93 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x94 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x95 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x96 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x97 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x98 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x99 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x9a */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x9b */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x9c */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x9d */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x9e */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */
	/* 0x9f */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},				/* SETcc */

	/* 0xa0 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* PUSH FS */
	/* 0xa1 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* POP FS */

	/* 0xa2 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* CPUID */

	/* 0xa3 */ {btx_handler, TYPE(LOAD)},                 /* BT reg, r/m */

	/* 0xa4 */ {todo},				/* SHLD IMM8 */
	/* 0xa5 */ {no_mem, INSTR_LEN(2), IMM_LEN(0)},				/* SHLD (CL) */

	/* 0xa6 */ {dontexist},
	/* 0xa7 */ {dontexist},

	/* 0xa8 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* PUSH GS */
	/* 0xa9 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				/* POP GS */

	/* 0xaa */ {privileged},				/* RSM */

	/* 0xab */ {btx_handler,TYPE(MODIFY)},                /* BTS reg, r/m */

	/* 0xac */ {no_mem, INSTR_LEN(3), IMM_LEN(1) /*todo */ },				/* SHRD IMM8 *///GOGO
	/* 0xad */ {no_mem, INSTR_LEN(2), IMM_LEN(0) /*todo */ },				/* SHRD (CL) *///GOGO

	/* 0xae */ {_sse},				/* FXRSTOR, FXSAVE, LDMXCSR, STMXCSR, SFENCE */

	/* 0xaf */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)},				/* IMUL reg,mem */

	/* 0xb0 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(MODIFY)},				/* CMPXCHG */
	/* 0xb1 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(MODIFY)},				/* CMPXCHG */

	/* 0xb2 */ {todo},				/* BTC */

	/* 0xb3 */ {btx_handler, TYPE(MODIFY)},               /* BTR reg, r/m*/

	/* 0xb4 */ {todo},				/* LFS */
	/* 0xb5 */ {todo},				/* LGS */

	/* 0xb6 */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(STORE)},	/* MOVZBL, MOVZBW met Poperand */
	/* 0xb7 */ {check_mod, IMM_LEN(0), MEM_LEN(2), TYPE(LOAD)},	/* MOVZWL */

	/* 0xb8 */ {dontexist},
	/* 0xb9 */ {dontexist},

	/* 0xba */ {todo},				/* BT, BTC, BTS */
	/* 0xbb */ {todo},				/* BTC */
	/* 0xbc */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)},				/* BSF -- TODO: correct memory instrumenting, it keeps scanning until it finds the first 1-bit !!!! */
	/* 0xbd */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(LOAD)},				/* BSR -- TODO: correct memory instrumenting, it keeps scanning until it finds the first 1-bit !!!! */

	/* 0xbe */ {check_mod, IMM_LEN(0), MEM_LEN(1), TYPE(LOAD)},	/* MOVSX, nooit een immediate! */
	/* 0xbf */ {check_mod, IMM_LEN(0), MEM_LEN(2), TYPE(LOAD)},	/* MOVSX */

	/* 0xc0 */ {todo},						/* XADD */
	/* 0xc1 */ {check_mod, IMM_LEN(0), MEM_LEN(4), TYPE(MODIFY)},						/* XADD */

	/* 0xc2 */ {_sse},						/* CMPPS */

	/* 0xc3 */ {dontexist},

	/* 0xc4 */ {_sse},				/* PEXTRW */
	/* 0xc5 */ {_sse},				/* PEXTRW */
	/* 0xc6 */ {_sse},				/* SHUFPS */

	/* 0xc7 */ {todo},				/* CMPXCHG8B */

	/* 0xc8 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				// bswap %eax
	/* 0xc9 */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				// bswap %ecx
	/* 0xca */ {no_mem, INSTR_LEN(1), IMM_LEN(0)},				// bswap %edx
	/* 0xcb */ {todo},
	/* 0xcc */ {todo},
	/* 0xcd */ {todo},
	/* 0xce */ {todo},
	/* 0xcf */ {todo},

	/* 0xd0 */ {_mmx},
	/* 0xd1 */ {_mmx},
	/* 0xd2 */ {_mmx},
	/* 0xd3 */ {_mmx},

	/* 0xd4 */ {dontexist},

	/* 0xd5 */ {_mmx},

	/* 0xd6 */ {dontexist},

	/* 0xd7 */ {_sse},				/* PMOVMSKB */

	/* 0xd8 */ {_mmx},
	/* 0xd9 */ {_mmx},
	/* 0xda */ {_mmx},				/* SSE: PMINUB!! */
	/* 0xdb */ {_mmx},
	/* 0xdc */ {_mmx},
	/* 0xdd */ {_mmx},
	/* 0xde */ {_mmx},				/* SSE: PMAXUB!! */
	/* 0xdf */ {_mmx},
	/* 0xe0 */ {_mmx},				/* SSE: PAVGB!! */
	/* 0xe1 */ {_mmx},
	/* 0xe2 */ {_mmx},
	/* 0xe3 */ {_mmx},				/* SSE: PAVGB!! */

	/* 0xe4 */ {_sse},				/* PMULHUW */

	/* 0xe5 */ {_mmx},

	/* 0xe6 */ {dontexist},

	/* 0xe7 */ {_sse},				/* MOVNTQ */

	/* 0xe8 */ {_mmx},
	/* 0xe9 */ {_mmx},
	/* 0xea */ {_mmx},				/* SSE: PWINSW!! */
	/* 0xeb */ {_mmx},
	/* 0xec */ {_mmx},
	/* 0xed */ {_mmx},
	/* 0xee */ {_mmx},				/* SSE: PMAXSW!! */
	/* 0xef */ {_mmx},
	/* 0xf0 */ {_mmx},
	/* 0xf1 */ {_mmx},
	/* 0xf2 */ {_mmx},
	/* 0xf3 */ {_mmx},

	/* 0xf4 */ {dontexist},

	/* 0xf5 */ {_mmx},

	/* 0xf6 */ {_sse},				/* PSADBW */
	/* 0xf7 */ {_sse},				/* MASKMOVQ */

	/* 0xf8 */ {_mmx},
	/* 0xf9 */ {_mmx},
	/* 0xfa */ {_mmx},
	/* 0xfb */ {_mmx},
	/* 0xfc */ {_mmx},
	/* 0xfd */ {_mmx},
	/* 0xfe */ {_mmx},
	/* 0xff */ {_mmx}
};

unsigned extended0F(unsigned dummy1, unsigned dummy2, unsigned dummy3){
	skip();
	return opcode0F_table[byte1].function(opcode0F_table[byte1].arg1,
			opcode0F_table[byte1].arg2,
			opcode0F_table[byte1].arg3);
}

/**********************************************************************************************/

unsigned LOOKUP_CLONE2INSTR(unsigned clone_address){
	unsigned result;

	// first check if there is a jump in the clone, if so return it
	if (GET_CHAR(clone_address) == 0xE9) {
		return clone_address+5+GET_LONG(clone_address+1);
	}
	LOCK_CLONE2INSTR();
	if (_clone2instr_[CLONE2INSTR_HASH(clone_address)].from
             == clone_address) {
		result = _clone2instr_[CLONE2INSTR_HASH(clone_address)].to;
		UNLOCK_CLONE2INSTR();
		return result;
	}
	UNLOCK_CLONE2INSTR();
	
	return 0;
}

/**********************************************************************************************/

void UPDATE_CLONE2INSTR(unsigned clone_address, unsigned X_instr_address) {
	// only update _clone2instr_ if there's no link from the clone to X_instr
	if (GET_CHAR(clone_address) != 0xE9) {
		LOCK_CLONE2INSTR();
		_clone2instr_[CLONE2INSTR_HASH(clone_address)].from = clone_address;
		_clone2instr_[CLONE2INSTR_HASH(clone_address)].to = (unsigned) X_instr_address;
		DBG(fprintf(jiti_tty, "\n>>>CLONE2INSTR: %x -> %x", 
					_clone2instr_[CLONE2INSTR_HASH(clone_address)].from,
					_clone2instr_[CLONE2INSTR_HASH(clone_address)].to));
		UNLOCK_CLONE2INSTR();
	}
}

/**********************************************************************************************/

void SCHEDULE_UPDATE_CLONE2INSTR(unsigned clone_address, unsigned X_instr_address) {
	if (_clone2instr_sched_count == CLONE2INSTR_SCHEDQUEUE) {
		panic("clone2instr schedule queue full!!!");
	}
	_scheduled_clone2instr_[_clone2instr_sched_count].from = clone_address;
	_scheduled_clone2instr_[_clone2instr_sched_count++].to = X_instr_address;
}

/**********************************************************************************************/

void COMMIT_CLONE2INSTR(void) {
	unsigned i;

	for (i = 0; i < _clone2instr_sched_count; i++) {
		UPDATE_CLONE2INSTR(_scheduled_clone2instr_[i].from,
		  _scheduled_clone2instr_[i].to);
	}
	_clone2instr_sched_count = 0;
}

/**********************************************************************************************/

void SCHEDULE_MAKE_LINK(unsigned from, unsigned to) {

	if (_make_link_sched_count == MAKE_LINK_SCHEDQUEUE) {
		panic("make_link schedule queue full!!!");
	}
	last_clone_link = from+5;
	DBG(fprintf(jiti_tty,"Last clone link is now 0x%x\n",last_clone_link));
	// put the address already in place
	PUT_LONG(from+1,to-from-5);
	// put the "jmp" opcode only there when the current code fragment is
	// done since otherwise another thread could start executing the
	// code we are currently generating before it's been finished
	_scheduled_make_link_[_make_link_sched_count++] = from;
}

/**********************************************************************************************/

void COMMIT_MAKE_LINK(void) {

	unsigned i;

	for (i = 0; i < _make_link_sched_count; i++) {
		PUT_CHAR(_scheduled_make_link_[i],0xE9);
	}
	_make_link_sched_count = 0;
}

/**********************************************************************************************/

unsigned LOOKUP_RELOC(unsigned address){
#ifndef USE_HASH
	int i;
#endif //USE_HASH
	unsigned result;

	if (!GEBRUIK_RELOC_INFO)
		return 0;

	LOCK_RELOC();
#ifdef USE_HASH
	DBG(fprintf(jiti_tty, "\n>>>[LOOKUP %x]", address));
	result = (_reloc_info_[HASH(address)]._from_ ==
			address) ? _reloc_info_[HASH(address)]._to_ : 0;
	UNLOCK_RELOC();
	return result;
#else
	DBG(fprintf(jiti_tty, "\n>>>[LOOKUP %x]", address));
	for (i = reloc_start_i; i != reloc_i; i = (i+1) % (RELOC)) {
		if (_reloc_info_[i]._from_ == address) {
			result = _reloc_info_[i]._to_;
			UNLOCK_RELOC();
			return result;
		}
	}
	DBG(fprintf(jiti_tty, "\n>>>NOT FOUND"));
	UNLOCK_RELOC();
	return 0;			/* niet gevonden */
#endif

}

/**********************************************************************************************/

unsigned decode_instruction(char show){
	X_kloon = 0;
	seg_override = 0;
	if (lock)
		panic("unprocessed lock!!!");

	X_fetch(X_address);
	return opcode_table[byte1].function(opcode_table[byte1].arg1,
			opcode_table[byte1].arg2,
			opcode_table[byte1].arg3);

}

/**********************************************************************************************/

#ifdef DEBUG
unsigned main_ret_address;
#endif //DEBUG

void jiti_restart_instr();

unsigned decode_block(unsigned address){
	// NOTE: do not initialize any of these variables in the declaration
	// with a value from a global variable, since such values can still
	// change before we get the lock in case of MT programs
	unsigned curdelta;
	unsigned X_instr_start;
	unsigned new_target;
	unsigned X_address_start;
	unsigned prev_end_jmp_target;
	unsigned old_patch_start;
	unsigned old_patch_end;
	DBG(unsigned X_address1);
	DBG(unsigned X_instr1);


	LOCK_DECODER();

	// OK, par -> we hebben de lock

	// init last_clone_link with current clone address
	last_clone_link = address;
	DBG(fprintf(jiti_tty,"decode_block start, last_clone_link = 0x%x\n",last_clone_link));


	old_patch_start = to_be_patched_start_i;
	old_patch_end = to_be_patched_i;

	if (address-delta > end)
		curdelta = ddelta;
	else
		curdelta = delta;
	address -= curdelta;

	// check whether the previous block of instrumented code ended with
	// a jump to the clone. If so, get the address of the position in the
	// clone this jump points to.
	prev_end_jmp_target = X_instr_end_jmp;
	if (prev_end_jmp_target)
		prev_end_jmp_target += 5+GET_LONG(X_instr_end_jmp+1);

	// check whether by the time we got the lock, we didn't already decode the
	// current block
	if ((new_target = LOOKUP_CLONE2INSTR(address+curdelta)) != 0) {
		if (prev_end_jmp_target == address+curdelta) {
			DBG(fprintf(jiti_tty,"Adjusting jump at 0x%x\n",X_instr_end_jmp);)
			MAKE_LINK(X_instr_end_jmp,new_target);
		}
		X_instr_end_jmp = 0;
		UNLOCK_DECODER();
		return new_target;
	}

	if ((new_target = LOOKUP_RELOC(address)) != 0) {
		if (prev_end_jmp_target == address+curdelta) {
			DBG(fprintf(jiti_tty,"Adjusting jump at 0x%x\n",X_instr_end_jmp);)
			MAKE_LINK(X_instr_end_jmp,new_target);
		}
		X_instr_end_jmp = 0;
		UNLOCK_DECODER();
		return new_target;
	}

	// check if we jump to the position in the clone corresponding with
	// the code we are going to instrument now at the end of the previous
	// X_instr code. If so, remove that jump and put the new instrumented
	// code right behind the previous.
	if (X_instr_end_jmp) {
		if (prev_end_jmp_target == address+curdelta) {
			DBG(fprintf(jiti_tty,"Removing jump at 0x%x\n",X_instr-5);)
			X_instr-=5;
		}
		X_instr_end_jmp = 0;
	}

	follow = 0;
	X_instr_start = (unsigned) X_instr;
	DBG(X_instr1 = (unsigned) X_instr);

#ifdef DEBUG
	// note: this may not be executed when we have several threads!!!
	if (tot_bp == 2) {
		fprintf(jiti_tty,"\n\nmain return address 0x%x is stored on the stack at 0x%x\n",main_ret,main_ret_address);
	}
#endif //DEBUG


	X_address = address;

	X_address_start = X_address;

	if (_instrument_code_) _X_start=X_address;

	// save the return address of main so that we know when we have to
	// stop instrumenting
	if (X_instr == X_instr_) {
#ifdef DEBUG
		PUT_SHORT(X_instr,0x2589);
		X_instr+=2;
		PUT_LONG(X_instr,&main_ret_address);
		X_instr+=4;  // movl %esp,main_ret_address
#endif //DEBUG
		PUSH_REG(EAX);
		X_instr++;
		PUT_LONG(X_instr,0x0424448b);
		X_instr+=4;  // movl 4(%esp),%eax
		MOVL_EAX_2_MEM(X_instr,&main_ret);
		X_instr+=5;
		POP_REG(EAX);
		X_instr++;
	}

	while (X_address) {

#ifdef DEBUG
		fprintf(jiti_tty, "\n### ");
		jiti_fprint_ins1(jiti_tty, 1, X_address);
		X_address1 = X_address;
		// update delta since when we follow direct jumps/calls, we can move
		// between the application and dynlib code
		if (X_address > end)
			curdelta = ddelta;
		else
			curdelta = delta;
#endif

		X_instr_start_block = (int) X_instr;
		if (GEBRUIK_RELOC_INFO) {
			UPDATE_RELOC(X_address, X_instr);
		}
		X_address = decode_instruction(1);

#ifdef DEBUG
		fprintf(jiti_tty, "\n*** ");
		jiti_fprint_ins1(jiti_tty, 1, X_address1 + curdelta);
		if (X_instr1 != X_instr) {
			fprintf(jiti_tty, "\n+++ ");
			jiti_fprint_ins(jiti_tty, X_instr1, X_instr);
		}
		X_instr1 = (unsigned)X_instr;
#endif

	}

	if (X_address_start > end)
		curdelta = ddelta;
	else
		curdelta = delta;

	// process scheduled MAKE_LINK operations. Do it before committing
	// the scheduled CLONE2INSTR entries, since addresses scheduled
	// for both will not be added to the _clone2instr_ table (they can
	// be looked up in the clone by checking whether the byte at the
	// address corresponding with the first byte of the original
	// instruction equals 0xe9)
	COMMIT_MAKE_LINK();
	// add scheduled clone2instr entries
	COMMIT_CLONE2INSTR();
	// update clone2instr information for this block
	UPDATE_CLONE2INSTR(X_address_start+curdelta,X_instr_start);

	if (GEBRUIK_PATCH_INFO) {
		unsigned i, new_target, last_not_patched = -1;
		unsigned start_was_smaller_than_end;
		DBG(fprintf(jiti_tty, "\n>>>TO BE PATCHED: %d ",
			(to_be_patched_i>=to_be_patched_start_i)?to_be_patched_i-to_be_patched_start_i:
			 to_be_patched_i+PATCH-to_be_patched_start_i));
		DBG(fprintf(jiti_tty,"\n>>>PATCH start index: %d, stop index: %d", to_be_patched_start_i, to_be_patched_i));
		for (i = to_be_patched_start_i; i != to_be_patched_i; i = (i+1) % PATCH) {
			DBG(fprintf(jiti_tty, "\n PATCH: %x:%x? ",
						_to_be_patched_info_[i]._address_,
						_to_be_patched_info_[i]._target_));
			if ((new_target = LOOKUP_CLONE2INSTR(_to_be_patched_info_[i]._target_+
					((_to_be_patched_info_[i]._target_ > end) ? ddelta:delta)))) {

				DBG(fprintf(jiti_tty, " PATCH: %x: %x -> %x (clone2instr)",
							_to_be_patched_info_[i]._address_,
							GET_LONG(_to_be_patched_info_[i]._address_) +
							_to_be_patched_info_[i]._address_,
							new_target));

				/*
				 * DBG(jiti_fprint_ins(jiti_tty, _to_be_patched_info_[i]._address_-2, 
				 * _to_be_patched_info_[i]._address_+4));
				 */
				if (i == to_be_patched_start_i) {
					to_be_patched_start_i = (to_be_patched_start_i + 1) % PATCH;
					if ((i == old_patch_start) &&
					    (old_patch_start != old_patch_end)) {
						old_patch_start = (old_patch_start + 1) % PATCH;
					}
				}

				if (!_to_be_patched_info_[i]._absolute_) {
					PUT_LONG(_to_be_patched_info_[i]._address_,
						new_target - _to_be_patched_info_[i]._address_ - 4);
				}
				else {
					PUT_LONG(_to_be_patched_info_[i]._address_,
						new_target);
				}

			} else if ((new_target=LOOKUP_RELOC(_to_be_patched_info_[i]._target_))) {

				DBG(fprintf(jiti_tty, " PATCH: %x: %x -> %x (reloc)",
							_to_be_patched_info_[i]._address_,
							GET_LONG(_to_be_patched_info_[i]._address_) +
							_to_be_patched_info_[i]._address_,
							new_target));

				if (i == to_be_patched_start_i) {
					to_be_patched_start_i = (to_be_patched_start_i + 1) % PATCH;
					if ((i == old_patch_start) &&
					    (old_patch_start != old_patch_end)) {
						old_patch_start = (old_patch_start + 1) % PATCH;
					}
				}
				//DBG(jiti_fprint_ins(jiti_tty, _to_be_patched_info_[i]._address_-2, _to_be_patched_info_[i]._address_+4));
				if (!_to_be_patched_info_[i]._absolute_) {
					PUT_LONG(_to_be_patched_info_[i]._address_,
						new_target - _to_be_patched_info_[i]._address_ - 4);
				}
				else {
					PUT_LONG(_to_be_patched_info_[i]._address_,
						new_target);
				}
			}
			else
				last_not_patched = i;
		}
		DBG(fprintf(jiti_tty,"\n>>>PATCH END before adjusting: start index: %d, stop index: %d", to_be_patched_start_i, to_be_patched_i));
		// remove the first half of the patch entries, since if they're not
		// yet patched now, they probably never will be using this technique
		// (they can still be patched inside jcc_taken_handler()).
		
		if (old_patch_start <= old_patch_end)
			to_be_patched_start_i += (old_patch_end - old_patch_start) / 2;
		else
			to_be_patched_start_i += (old_patch_end + PATCH - old_patch_start) / 2;
		to_be_patched_start_i = to_be_patched_start_i % PATCH;

		// if we patched some entries at the end of the current patch info array,
		// don't try to patch them next time
		start_was_smaller_than_end = to_be_patched_start_i <= to_be_patched_i;
		if (last_not_patched != -1)
			to_be_patched_i = (last_not_patched+1) % PATCH;
		if (start_was_smaller_than_end &&
		    (to_be_patched_i < to_be_patched_start_i))
			to_be_patched_i = to_be_patched_start_i;

		DBG(fprintf(jiti_tty,"\n>>>PATCH END after adjusting: start index: %d, stop index: %d", to_be_patched_start_i, to_be_patched_i));
	}

	UNLOCK_DECODER();

	return X_instr_start;

}

/********************************************************************************************/

#define INITIAL_PAR_SIZE 128*1024*1024
#define INITIAL_SEQ_SIZE 32*1024*1024	// minimaal 8kB !!
#define REDZONE_FACTOR   4
unsigned x_redzone = 0;

time_t time_of_last_realloc = 0;

/////

void check_x_instr_size() {
	// we lopen in REDZONE => opkuisen en herstarten
	if ((X_instr + x_redzone) > (X_instr_ + x_instr_size))
		jiti_restart_instr();
}

////

void jiti_start_instr(){

	time(&time_of_last_realloc);

	x_instr_size =
		jiti_parallel_program()? INITIAL_PAR_SIZE : INITIAL_SEQ_SIZE;
	x_redzone = x_instr_size / REDZONE_FACTOR;

	if (!(X_instr_ = (char *) malloc(x_instr_size))) panic("malloc() failed\n");

	DBG(fprintf(jiti_tty, "X_instr_: Allocating %d bytes from 0x%x to 0x%x @ %d, "
				"REDZONE (%d bytes) starts at 0x%x\n", x_instr_size,
				(unsigned) X_instr_, (unsigned) X_instr_ + x_instr_size,
				(unsigned) time_of_last_realloc, x_redzone,
				(unsigned) X_instr_ + x_redzone));
	X_instr = X_instr_;
}

/////

void jiti_restart_instr(){
	time_t new_time;

	time(&new_time);

	DBG(fprintf
			(jiti_tty, "X_instr_: RESTARTING @ %d\n", (unsigned) new_time));

	fprintf(stderr, "X_instr_: RESTARTING @ %d\n", (unsigned) new_time);

	if (jiti_parallel_program())
		panic("Reinstrumenting parallel program!\n");

	// seq. program

	DBG(fprintf
			(jiti_tty, "X_instr_: Freeing    %d bytes @ 0x%x\n", x_instr_size,
			 (unsigned) X_instr_));
	free(X_instr_);

	// nieuwe grootte voor X_instr_ bepalen
	{
#define MAKE_SMALLER 60
#define MAKE_BIGGER  10

		/*
		 * MAKE_BIGGER                        MAKE_SMALLER
		 * --------------+----------------------------------+---------------------->
		 * 
		 * *2                      =                            /2
		 * 
		 */

		if (new_time > (time_of_last_realloc + MAKE_SMALLER)) {
			x_instr_size >>= 1;	// /2
		} else if (new_time < (time_of_last_realloc + MAKE_BIGGER)) {
			x_instr_size <<= 1;	// *2
		}
		x_redzone = x_instr_size / REDZONE_FACTOR;
		time_of_last_realloc = new_time;
	}

	// clone2instr invalideren
	memset(_clone2instr_,0,sizeof(_clone2instr_));

	// _reloc_info_ invalideren
	{
		unsigned i;
		for (i = 0; i < RELOC; i++) {
			_reloc_info_[i]._from_ = 0;
			_reloc_info_[i]._to_ = 0;
		}
	}

	// _to_be_patched_info_ invalideren
	{
		unsigned i;
		for (i = 0; i < PATCH; i++) {
			_to_be_patched_info_[i]._address_ = 0;
			_to_be_patched_info_[i]._target_ = 0;
		}
	}

	// nieuwe X_instr_ aanmaken
	if (!(X_instr_ = (char *) malloc(x_instr_size)))
		panic("malloc() failed\n");
	DBG(fprintf
			(jiti_tty,
			 "X_instr_: Allocating %d bytes @ 0x%x REDZONE (%d bytes) starts at 0x%x\n",
			 x_instr_size, (unsigned) X_instr_, x_redzone,
			 (unsigned) X_instr_ + x_redzone));
	X_instr = X_instr_;
}
