/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  jiti.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: jmaebe $
 * $Date: 2002/03/26 13:15:52 $
 * $Log: jiti.c,v $
 * Revision 1.9.2.16  2002/03/26 13:15:52  jmaebe
 * * fixed detection of PLT jmp_mem entries
 * * if we detect that we are going to return to a breakpoint in
 *   LOOKUP_RET_ADDRESS, we immediately instrument that code
 * * the jump coming after instrumented "int 0x80"'s is now replcaced with a
 *   a call to jmpcall_dir_handler too
 * * the clone wasn't refilled partially with breakpoints when we followed a
 *   direct short jump (with instruction length 2)
 *  -> with the 3 previous changes, we now only touch breakpoints when starting
 *   the program and when returning from the first call to a routine in
 *   libpthread (-> FPC has only one breakpoint left, pdf2ps 7)
 *
 * Revision 1.9.2.15  2002/03/15 11:26:43  jmaebe
 * - disabled use of lookup_reloc in lookup_ret_handler because lookup_reloc
 *   is not thread safe and more importantly, never seems to produce a match
 *   there (probably because it only contains very local information)
 *
 * Revision 1.9.2.14  2002/03/14 17:29:26  jmaebe
 * * optimized the hell out of the reloc/patch facilities -> on moore a gzip
 *   operation is only 20% slower with the opt backend than without!!
 *
 * Revision 1.9.2.13  2002/03/11 21:21:59  jmaebe
 * * MAKE_LINK is now no longer used while generating code, since it also
 *   caused problems when using MT code (thread ! is instrumenting, generats a
 *   link to its instrumented code which isn't finished yet, thread 2 returns
 *   from a function or jumps/calls to the linked address and starts executing
 *   the unfinished instrumented code) -> changed to SCHEDULE_MAKE_LINK() and
 *   committed after the current block is finished
 * * LOOKUP_CLONE2INSTR() once again uses the jumps in the clone to determine
 *   targets
 * * libpthread is no longer instrumented because it disables all signals
 *   (using syscall 0xb3) in some routines, causing JiTI to hang (since its
 *   breakpoints aren't processed anymore, it can't continue instrumenting and
 *   the program can't continue executing)
 * * made jcc_taken_handler() thread-safe: the MAKE_LINK() there was not safe
 *   since it consists of 2 PUT_****() operations, so another thread could be
 *   executing the original call while it was being replaced with a JMP. At
 *   the same time, PATCH_INFO was made obsolete by doing the patching inside
 *   jcc_taken_handler (and that should be more or less thread safe, since it
 *   replaces exactly one int) -> GEBRUIK_PATCH_INFO is always set to 0 in
 *   jiti.c, can possibly be removed.
 * + infrastructure for intercepting syscalls and modifying the values passed
 *   to the kernel, not used anymore (I used it to debug the syscall 0xb3
 *   problem)
 * * separate assembler handlers for adjusting the target of a jmp or call
 *   mem/reg when instrumenting parallel programs (the call handler checks
 *   for targets in libpthread, the jump handler doesn't). The common handler
 *   is still used if we aren't instrumenting a parallel program (and that
 *   one obviously doesn't include the pthread checks)
 * * the memory layout is now always written to the log
 * - commented out some unused global variables in instrument.c
 * * some cosmetic changes and translations
 * * with the above changes, ploop now always works (on a dual proc machine)
 *   with the opt and trace-mem backends). ptest now always crashed for some
 *   reason however. If you use trace-code, you also almost always get crashes
 *   with MT programs.
 * And that's all for today, folks!
 *
 * Revision 1.9.2.12  2002/03/06 20:54:55  jmaebe
 * * sped up jiti by about 1000% by using a hashing function to lookup entries
 *   in clone2instr. FPC now runs only about 4 times slower with instrumentation
 *   than without.
 *
 * Revision 1.9.2.11  2002/03/04 15:02:22  jmaebe
 * - removed now unused global bp variable
 *
 * Revision 1.9.2.10  2002/02/28 15:53:34  jmaebe
 * * signal handlers are re-entrant under Linux nowadays, but bp_handlet set/
 *   used a global variable, causing general havoc with multi-threaded programs
 *   -> bp changed to a local variable (it isn't used outside bp_handler anyway)
 * * __test_and_set() prototype moved from insturment.c to jiti.h
 * * clone2instr updating now happens after the current block has been finished,
 *   otherwise a thread could start executing the unfinished block of
 *   instructions that another thread is generating
 * * interrupt_handler() only sets X_kloon to 1 if we aren't instrumenting code
 * * removed unnecessary locks in jcc_taken_handler()
 *
 * !!WARNING: don't ever use pthread_mutex_lock() etc, because these functions
 *   don't seem to be re-entrant so we hang if we call them while we are
 *   instrumenting them.
 *
 * Revision 1.9.2.9  2002/02/27 15:03:14  jmaebe
 * * made RELOC and CLONE2INSTR thread-safe
 * * fixed use of UPDATE_RELOC macro
 * * moved LOOKUP_CLONE2INSTR from jiti.c to instrument.c
 * * fixed usage of LOOKUP_CLONE2INSTR (in most cases, an address in the
 *   original program instead of in the clone was passed, so it never found
 *   anything)
 *
 * Revision 1.9.2.8  2002/02/25 16:25:17  jmaebe
 * + deactivate jiti when exit() is called too (fixes crash at the end of gzip)
 *
 * Revision 1.9.2.7  2002/02/22 17:03:42  jmaebe
 * * made checking of ret from main thread safe
 * * LOOKUP_RET_HANDLER now also uses LOOKUP_CLONE2INSTR and LOOKUP_RELOC when
 *   the ret address is in a dynlib
 * * ret_handler (handlers.s) now saves all registers (it calls the C-routine
 *   LOOKUP_RET_HANDLER), should be slightly faster and is now jumped at
 *   instead of called
 * * fac now works reliably to when instrumented!!
 *
 * Revision 1.9.2.6  2002/02/22 14:23:31  jmaebe
 * + save return address of main at the start of the program and use it to
 *   stop instrumenting when we return from main
 *
 * Revision 1.9.2.5  2002/02/21 14:27:23  jmaebe
 * * the gathering of the memory regions occupied by dynlibs is now done from
 *   /proc/self/maps (like in jiti-malloc) -> ld.so is now also found.
 *
 * Revision 1.9.2.4  2002/02/20 12:23:20  jmaebe
 * - removed all "ifdef INSTRUMENT_DYNLIBS" directives (always enabled now)
 *
 * Revision 1.9.2.3  2002/02/14 17:43:40  jmaebe
 * + Support for the Free Pascal Compiler added (merged from main branch)
 *
 * Revision 1.9.2.2  2002/02/13 12:16:59  jmaebe
 * + support for interrupts
 * * instrumenting dynamic libraries now works, but libc still causes some
 *   problems because of unsupported instructions (e.g lock and cmpxchgl)
 *
 * Revision 1.9.2.1  2002/02/12 16:47:20  jmaebe
 * + initial implementation of instrumenting of dynamically loaded libraries,
 *   but not yet working (even main(void) {} crashes :). Also, some backends
 *   which have to be modified before they can work with this new feature
 *   haven't been adapted yet. Compiler with -DINSTRUMENT_DYNLIBS to include
 *   the functionality (it's even required currently, because there's a
 *   reference to ddelta (dynlib delta) in handlers.s which can't be disabled
 *   using a conditional. Method used: separate clone for the dynlibs (thanks
 *   for the idea, Michiel :)
 *
 * Revision 1.9  2002/01/15 17:26:40  ronsse
 * Added empty line to test cvs
 *
 * Revision 1.8  2002/01/10 21:25:30  ronsse
 * *** empty log message ***
 *
 * Revision 1.7  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


//#define DEBUG

 

#include <stdio.h>
#include <unistd.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <signal.h>      
#include <dlfcn.h> 

#include "jiti.h"

#include "dynlib.h"


#define BP 0xCC   /* int 3 */

FILE * jiti_tty = NULL ;

unsigned _elf_start;  // adres van _start
unsigned _elf_init;   // adres van _init
unsigned _elf_fini;   // adres van _fini

int _jiti_par_exec=0, _jiti_par_program=0;

unsigned _optimize_=0;

char * __clone__;
unsigned delta;

char * __dynlibclone__;
// ddelta is the distance between the clone for the dynlibs and the
// dynlibs themselves
unsigned ddelta;


unsigned begin=0x08048000;
unsigned end;

void pushl_imm_handler(void);
void pushl_imm8_handler(void);

void call_imm_handler(void);
void ret_handler(void);
void leave_handler(void);

extern unsigned GEBRUIK_RELOC_INFO;
extern unsigned GEBRUIK_PATCH_INFO;
extern unsigned FOLLOW_JMP_DIR;
extern unsigned FOLLOW_CALL_DIR;

extern unsigned X_instr_,X_instr;
extern unsigned __instr_p__;

unsigned main_ret;
unsigned addr;

unsigned tot_bp=0;

unsigned _backend_access_ok_;

// if the program calls exit(), we have to make sure we stop instrumenting
// when this happens because jiti will cash when the fini routines of the
// libs it depends on execute
unsigned exit_jmpaddr;

char _jiti_naam[100];

unsigned _main_=0;

// callbacks
void CB_empty(){}  // empty default callback
void ___internal___error___002(){panic("___internal___error___#002");}

unsigned CB_data  = (unsigned)___internal___error___002;
unsigned CB_code  = (unsigned)___internal___error___002;
unsigned CB_stop  = (unsigned)CB_empty;
unsigned CB_start = (unsigned)CB_empty;


/****************************************************************************/

void panic(char * string){
  fprintf(jiti_tty, "\n\n>> PANIC: %s <<\n\n\n", string);
  printf("\n\n>> PANIC: %s <<\n\n\n", string);
  _exit(-1);
}

/****************************************************************************/

void check_x_instr_size();

void bp_handler(int par){
  unsigned bp;
  unsigned addr;
  unsigned new_target;
#ifdef DEBUG
  unsigned len;
#endif

#define JUMP_TO(addr) \
	PUT_LONG(bp+68, addr);


  bp   = get_bp();
  addr = GET_LONG(bp+68)-1;

  tot_bp++;


  // is er nog plaats voor de nieuwe code??
  check_x_instr_size();

#ifdef DEBUG  
  fprintf(jiti_tty, "\n\n<<<BP: ");
  if (addr-delta > end)
    len=jiti_fprint_ins1(jiti_tty,1,addr-ddelta);
  else
    len=jiti_fprint_ins1(jiti_tty, 1, addr-delta);
  //  jiti_fprint_ins1(jiti_tty, 1, addr);
  fprintf(jiti_tty, "#%d>>>", len);
#endif

  if (((new_target=LOOKUP_CLONE2INSTR(addr)) != 0) ||
      (GET_CHAR((new_target = addr)) != BP)) {
    DBG(fprintf(jiti_tty, "\n>>>--Already instrumented, jumping to %x", new_target));
  } else {
    new_target=decode_block(addr);
    FLUSH_INSTRUCTIONS;
    DBG(fprintf(jiti_tty, "\n>>>--Restarting, jumping to %x", new_target));
  }

  JUMP_TO(new_target);

}   
/****************************************************************************/


unsigned ori_start, ori_stop, clone_start, clone_stop;

unsigned plt_end;

/********************************************************/

#include <stdlib.h>      

int _jiti_init2(int argc, char ** argv);
int _jiti_init1();

/* Deze init-functie wordt opgeroepen door ELKE backend */
int jiti_init(unsigned backend_callback){
  static int _jiti_init_called = FALSE;

  // EEnmalige initialisatie van JiTI m.b.v. _jiti_init1()
  if (!_jiti_init_called) {
    if (_jiti_init1()) return -1;   // geen ./prog
    _jiti_init_called = TRUE;
  }

  // vanaf nu kunnen we jiti_tty gebruiken voor de logboodschappen

  DBG(fprintf(jiti_tty, "===== JiTI_init: backend_callback @ 0x%x / start\n", backend_callback));

  _backend_access_ok_=1;
  CALL_FUNCTION(backend_callback);
  _backend_access_ok_=0;  

  DBG(fprintf(jiti_tty, "===== JiTI_init: backend_callback @ 0x%x / stop\n", backend_callback));

  return 0;
}


/***************************************************************/

int fini1();

/* Dit is DE _init functie, ze wordt slechts EENMAAL opgeroepen door jiti_init() */
int _jiti_init1(){
  int pid;
  unsigned call_main;
  FILE * file;
  
  /* check for local file */
  pid=getpid();
  sprintf(_jiti_naam, "/proc/%d/cmdline", pid);
  if (!(file=fopen(_jiti_naam, "r"))) perror(_jiti_naam);
  fscanf(file, "%s", _jiti_naam);
  fclose(file);

  /* '/' als start van de naam => return */
  if (*_jiti_naam=='/') return -1;

  /* niet in . => return */
  if (!(file=fopen(_jiti_naam, "r"))) return -1;
  fclose(file);


  /********* OK, Let's do it!  *********/


  /* open uitvoer */
  {    
    char log_name[100];
    int i;
    sprintf(log_name, "jiti.log-%s", _jiti_naam);
    for (i=0; i<strlen(log_name); i++)
      if (log_name[i]=='/')
        log_name[i]='_';
    jiti_tty = fopen(log_name, "w");
    if ((int)jiti_tty==-1) jiti_tty=stderr;
    setvbuf(jiti_tty, NULL, _IONBF, 0);
    fprintf(jiti_tty, "===== %s\n===== CMD: %s\n", &jiti_version, _jiti_naam);
  }

  /* lees ELF-header en bouw PLT-tabel op*/
  process_elf_info(_jiti_naam);

  /* bepaal startadres */
  call_main = _elf_start;
  plt_end   = call_main;

  DBG(fprintf(jiti_tty, "\n===== JiTI_init1: start P%d [%s]\n", pid, _jiti_naam));

  /* registeer LAATSTE fini */
  atexit((void(*)(void))fini1);


  /* initialiseer eigen backends, gebeurt vOOr alle andere backend-initialisaties!!! */
  _backend_access_ok_=1;
  par_init();                          // initialiseer par()-backend
  _backend_access_ok_=0;


  // Check LD_BIND_NOW??


  /* CALL MAIN (mogelijks op twee paginas) moet ook schrijfbaar zijn */  
  if (mprotect ( (void*)((call_main/4096)*4096), 4096*2, PROT_EXEC|PROT_READ|PROT_WRITE)==-1) {
    perror("mprotect failed");
    _exit(-1);
  }
  

  // CALL/PUSH MAIN wijzigen in CALL _JITI_INIT2

  switch (*(unsigned*)call_main)  { 
    
  case 0xd285ed31:

    // gcc version 2.95.1: CALL MAIN @ call_main+0x36
    call_main += 0x36;

    // _main_ is het *absolute* doeladres
    _main_ = GET_LONG(call_main)+call_main+4;

    // plaat *relatieve* adres van _jiti_init2() in de plaats
    PUT_LONG(call_main, _jiti_init2-call_main-4);

    break;

  case 0x895eed31:    
    
    // egcs-2.91.66: PUSHL ABS @ call_main+0x18
    call_main += 0x18; 

    // _main_ is het *absolute* doeladres
    _main_ = GET_LONG(call_main);
    
    // plaat *absolute* adres van _jiti_init2() in de plaats
    PUT_LONG(call_main, _jiti_init2);

    break; 
    
  case 0x89c7895e:

    // fpc
    call_main+=51;

    // _main_ is het *absolute* doeladres
    _main_ = GET_LONG(call_main);

    fprintf(jiti_tty, "===== @main @ 0x%x, main @ 0x%x\n", call_main, _main_);

    // plaat *absolute* adres van _jiti_init2() in de plaats
    PUT_LONG(call_main, _jiti_init2);
    break; 

  default:
    panic("Unknown gcc/egcs!\n");
    break;
    
  }

  // get dynlib info (base/size)
  get_dynlibs_info();

  DBG(fprintf(jiti_tty, "===== @main @ 0x%x, main @ 0x%x\n", call_main, _main_));

  DBG(fprintf(jiti_tty, "===== JiTI_init1: done\n"));
  return 0;
}

/*******************************************************/

int fini2();

void jiti_start_instr(); // instrument.c

int _jiti_init2(int argc, char ** argv){


  // op dit moment weten we reeds of het een seq/par programma is. Te checken
  // m.b.v. jiti_parallel_program().

#if 1
  fprintf(jiti_tty, "===== JiTI_init2: start\n");
  fprintf(jiti_tty, "===== Instrumenting: ");
  if (_instrument_mem_)  fprintf(jiti_tty, "MEM ");
  if (_instrument_code_) fprintf(jiti_tty, "CODE ");
  if (_instrument_dyn_)  fprintf(jiti_tty, "DYN ");
  fprintf(jiti_tty, "\n");
#endif

  atexit((void(*)(void))fini2);

  // Bepaal de gewenste instrumentatie
  ///// Bepaal optimalisatie
  
  // Er is (voorlopig) slechts 1 regel: relocatie en patchen is niet toegestaan indien we code tracen
  if (_instrument_code_){
    GEBRUIK_RELOC_INFO=0;
    GEBRUIK_PATCH_INFO=0;
    FOLLOW_JMP_DIR=0;
    FOLLOW_CALL_DIR=0;
  }

  // Het gebruik van PATCH_INFO vereist het bestaan van RELOC_INFO
  if (GEBRUIK_PATCH_INFO) GEBRUIK_RELOC_INFO=1;


  /* Initialiseer info voor disassemble() */
  _jiti_init_info();


  if (_instrument_mem_ || _instrument_code_ || _optimize_) {
    
    // we instrumenteren data en/of code
    
    /* Bepaal grootte origineel */
    free((void*)(end=(int)malloc(1)));
    DBG(fprintf(jiti_tty, "Code: [0x%.8x - 0x%.8x]\n", begin, end));
    
    /* Allokeer kloon */
    __clone__=(char*)malloc(end-begin);
    
    __dynlibclone__=(char*)malloc(_dynlibs_size);

    /* Bepaal delta */
    delta=(char*)__clone__-(char*)begin;
    DBG(fprintf(jiti_tty, "Clone [0x%.8x - 0x%.8x] length=%d, delta=%d\n", 
      (unsigned)__clone__, (unsigned)__clone__+end-begin, end-begin, delta)); 
    
    ddelta=(char*)__dynlibclone__-(char*)_dynlibs_base;
    DBG(fprintf(jiti_tty, "Dynlib clone [0x%.8x - 0x%.8x] length=%d, ddelta=%d\n", 
      (unsigned)__dynlibclone__, (unsigned)__dynlibclone__+_dynlibs_size, _dynlibs_size, ddelta)); 


    ori_start   = begin;
    ori_stop    = end;
    clone_start = (unsigned)__clone__;
    clone_stop  = (unsigned)__clone__+end-begin;    

    /* vul kloon met breekpunten */
    memset(__clone__, BP, end-begin);

    memset(__dynlibclone__, BP, _dynlibs_size);
    
    /* installeer BP-handler */
    signal(SIGTRAP, bp_handler); 
    
    // allokeer plaats voor de geinstrumenteerde code
    jiti_start_instr();

    // Zorg dat we `delta` verder springen
    _main_ += delta;
    
  }

  // make sure we stop instrumenting when exit() is called
  exit_jmpaddr = get_exit_jmpaddr();

  // roep nu main(+delta) op, met de juiste argumenten
  DBG(fprintf(jiti_tty, "===== JiTI_init2: done\n===== starting main...\n"));

  _backend_access_ok_=1;
  CALL_FUNCTION(CB_start);
  _backend_access_ok_=0;
  
  return (*((int(*)(int, char**))(_main_)))(argc, argv);
  
}

/*******************************************************/


int fini2(){

  DBG(fprintf(jiti_tty, "===== JiTI_fini2\n"));

  // ...

  DBG(fprintf(jiti_tty, "===== Backend fini()s:\n"));

  return 0;
}

/*******************************************************/

int fini1(){
  int i=0,j=0;

  // tel het aantal aangemaakte stukjes kloon
  while (j < CLONE2INSTR_SIZE)
	i+= (_clone2instr_[j++].from != 0);

#if 1
  fprintf(jiti_tty, "===== JiTI_fini1\n");
  fprintf(jiti_tty, "===== Used %d bytes instrumentation code\n", (unsigned)X_instr-(unsigned)X_instr_);
  fprintf(jiti_tty, "===== Breakpoints: %d in cache, %d executions\n", i, tot_bp);
  fprintf(jiti_tty, "===== Backend fini()s:\n\n\n");
#endif

  CALL_FUNCTION(CB_stop);

  fclose(jiti_tty);

  // THE END

  return 0;
}

/*******************************************************/

int jiti_parallel_execution() { return _jiti_par_exec;    }
int jiti_parallel_program()   { return _jiti_par_program; }

/*******************************************************/

unsigned LOOKUP_RET_ADDRESS(unsigned clone_address){
  unsigned curdelta;
  unsigned address;

  if (clone_address > end)
    curdelta=ddelta;
  else
    curdelta=delta;
  clone_address+=curdelta;

  address=LOOKUP_CLONE2INSTR(clone_address);

  DBG(fprintf(jiti_tty, "\n>>> RET: LOOKUP_CLONE2INSTR(%x) -> %x", clone_address, address));

  if (address) return address;

/* this (almost?) never produces a match (reloc info is very local) and is
   currently also not thread-safe (nothing is locked while the reloc info
   gets updated -> disabled
  address=LOOKUP_RELOC(clone_address-curdelta);

  DBG(fprintf(jiti_tty, "\n>>> RET: LOOKUP_RELOC(%x) -> %x", clone_address-curdelta, address));

  if (address) return address;
*/

  if (GET_CHAR(clone_address) == BP) {
        DBG(fprintf(jiti_tty,"\n>>> RET: code in clone is a breakpoint -> instrumenting..."));
	clone_address = decode_block(clone_address);
	FLUSH_INSTRUCTIONS;
  }
  else {
    DBG(fprintf(jiti_tty,"\n>>> RET: code in clone (0x%x) is valid, doing nothing - ",clone_address));
    DBG(jiti_fprint_ins1(jiti_tty,1,clone_address));
  }

  return clone_address;

}

/////////////////////////////////////////////////////

void optimize_execution(){
  _optimize_=1;
}

char * jiti_program_name(){ 
  return _jiti_naam;
}


